# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

""" Code adopted from https://github.com/MilkaLichtblau/xing_dataset/blob/master/src/XingProfilesReader.py """

import os
import sys
import uuid
import glob
import math
import json
import datetime
import pandas as pd
from typing import Sequence, Union

from ._base import TabDataset


class Candidate():
    """
    represents a candidate in a set that is passed to a search algorithm
    a candidate composes of a qualification and a list of protected attributes (strings)
    if the list of protected attributes is empty/null this is a candidate from a non-protected group
    natural ordering established by the qualification
    """

    def __init__(self, qualification, protectedAttributes):
        """
        @param qualification : describes how qualified the candidate is to match the search query
        @param protectedAttributes: list of strings that represent the protected attributes this
                                    candidate has (e.g. gender, race, etc)
                                    if the list is empty/null this is a candidate from a non-protected group
        """
        self.__qualification = qualification
        self.__protectedAttributes = protectedAttributes
        # keeps the candidate's initial qualification for evaluation purposes
        self.__originalQualification = qualification
        self.uuid = uuid.uuid4()

    @property
    def qualification(self):
        return self.__qualification

    @qualification.setter
    def qualification(self, value):
        self.__qualification = value

    @property
    def originalQualification(self):
        return self.__originalQualification

    @originalQualification.setter
    def originalQualification(self, value):
        self.__qualification = value

    @property
    def isProtected(self):
        '''
        true if the list of ProtectedAttribute elements actually contains anything
        false otherwise
        '''
        return not self.__protectedAttributes == []


class Xing(TabDataset):
    """ https://github.com/MilkaLichtblau/xing_dataset """

    file2url = {
        "SHAano01_1-40.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano01_1-40.json",
        "SHAano02-41-80.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano02-41-80.json",
        "SHAano03_81-120.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano03_81-120.json",
        "SHAano04_121-160.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano04_121-160.json",
        "SHAano05_161-200.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano05_161-200.json",
        "SHAano06_201-240.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano06_201-240.json",
        "SHAano07_241-280.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano07_241-280.json",
        "SHAano08_281-320.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano08_281-320.json",
        "SHAano09_321-360.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano09_321-360.json",
        "SHAano10_361-400.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano10_361-400.json",
        "SHAano11_401-440.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano11_401-440.json",
        "SHAano12_441-480.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano12_441-480.json",
        "SHAano13_481-520.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano13_481-520.json",
        "SHAano14_521-560.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano14_521-560.json",
        "SHAano15_561-600.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano15_561-600.json",
        "SHAano16_601-640.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano16_601-640.json",
        "SHAano17-641-680.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano17-641-680.json",
        "SHAano18_681-720.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano18_681-720.json",
        "SHAano19_721-760.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano19_721-760.json",
        "SHAano20_761-800.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano20_761-800.json",
        "SHAano21_801-840.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano21_801-840.json",
        "SHAano22_841-880.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano22_841-880.json",
        "SHAano23_881-920.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano23_881-920.json",
        "SHAano24_921-960.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano24_921-960.json",
        "SHAano25_961-1000.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano25_961-1000.json",
        "SHAano26_1001-1040.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano26_1001-1040.json",
        "SHAano27_1041-1080.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano27_1041-1080.json",
        "SHAano28_1081-1120.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano28_1081-1120.json",
        "SHAano29_1121-1160.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano29_1121-1160.json",
        "SHAano30_1161-1200.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano30_1161-1200.json",
        "SHAano31_1201-1240.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano31_1201-1240.json",
        "SHAano32_1241-1280.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano32_1241-1280.json",
        "SHAano33_1281-1320.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano33_1281-1320.json",
        "SHAano34_1320-1360.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano34_1320-1360.json",
        "SHAano35-1361-1400.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano35-1361-1400.json",
        "SHAano36-1401-1440.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano36-1401-1440.json",
        "SHAano37_1441-1480.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano37_1441-1480.json",
        "SHAano38_1481-1520.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano38_1481-1520.json",
        "SHAano39_1521-1560.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano39_1521-1560.json",
        "SHAano40_1561-1600.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano40_1561-1600.json",
        "SHAano41_1601-1640.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano41_1601-1640.json",
        "SHAano42_1641-1680.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano42_1641-1680.json",
        "SHAano43_1681-1720.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano43_1681-1720.json",
        "SHAano44_1721-1760.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano44_1721-1760.json",
        "SHAano45_1760-1800.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano45_1760-1800.json",
        "SHAano46_1801-1840.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano46_1801-1840.json",
        "SHAano47_1841-1880.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano47_1841-1880.json",
        "SHAano48_1881-1920.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano48_1881-1920.json",
        "SHAano49_1921-1960.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano49_1921-1960.json",
        "SHAano50_1961-2000.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano50_1961-2000.json",
        "SHAano51_2001-2040.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano51_2001-2040.json",
        "SHAano52_2041-2080.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano52_2041-2080.json",
        "SHAano53_2081-2120.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano53_2081-2120.json",
        "SHAano54_2121-2160.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano54_2121-2160.json",
        "SHAano55_2161-2200.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano55_2161-2200.json",
        "SHAano56_2201-2240.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano56_2201-2240.json",
        "SHAano57_2241-2280.json": "https://raw.githubusercontent.com/MilkaLichtblau/xing_dataset/master/data/SHAano57_2241-2280.json",
    }

    EDUCATION_OR_JOB_WITH_NO_DATES = 3  # months count if you had a job that has no associated dates
    EDUCATION_OR_JOB_WITH_SAME_YEAR = 6  # months count if you had a job that started and finished in the same year
    EDUCATION_OR_JOB_WITH_UNDEFINED_DATES = 1  # month given that the person entered the job

    def __init__(
            self,
            sen_feat: Union[str, Sequence[str]] = "sex",
            dir_path: str = os.path.join(sys.path[1], "fairpy/data/xing"),
            download: bool = False,
    ):
        self._check_and_download_data(dir_path, self.file2url, download)

        self.entireDataSet = pd.DataFrame(columns=['protected', 'nonProtected', 'originalOrdering'])

        files = glob.glob(dir_path + "/*.json")

        for filename in files:
            print(filename)
            key, protected, nonProtected, origOrder = self.__readFileOfQuery(filename)
            self.entireDataSet.loc[key] = [protected, nonProtected, origOrder]

        print(self.entireDataSet["originalOrdering"]["Dental Assistant"][20].qualification)

    def __readFileOfQuery(self, filename):
        """
        takes one .json file and reads all information, creates candidate objects from these
        information and sorts them into 3 arrays. One contains all protected candidates, one contains
        all non-protected candidates, one contains all candidates in the same order as they appear
        in the json-file
        @param filename: the json's filename
        @return:
            key: the search query string
            protected: array that contains all protected candidates
            nonProtected: array that contains all nonProtected candidates
        """

        protected = []
        nonProtected = []
        originalOrdering = []

        currentfile = open(filename)
        data = json.load(currentfile)

        xingSearchQuery = data['category']
        # if the Xing search query results in a gender neutral list,
        # we take female as the protected attribute
        protectedAttribute = 'm' if data['dominantSexXing'] == 'f' else 'f'

        for r in data['profiles']:
            # determine Member since / Hits
            if 'memberSince_Hits' in r['profile'][0]:
                hits_string = r['profile'][0]['memberSince_Hits']
                hits = hits_string.split(' / ')[1]
            else:
                hits = 1

            work_experience = self.__determineWorkMonths(r)
            edu_experience = self.__determineEduMonths(r)
            score = (work_experience + edu_experience) * int(hits)

            if self.__determineIfProtected(r, protectedAttribute):
                protected.append(Candidate(score, [protectedAttribute]))
                originalOrdering.append(Candidate(score, [protectedAttribute]))
            else:
                nonProtected.append(Candidate(score, []))
                originalOrdering.append(Candidate(score, []))

        protected.sort(key=lambda candidate: candidate.qualification, reverse=True)
        nonProtected.sort(key=lambda candidate: candidate.qualification, reverse=True)

        self.__normalizeQualifications(protected + nonProtected)
        self.__normalizeQualifications(originalOrdering)

        currentfile.close()

        return xingSearchQuery, protected, nonProtected, originalOrdering

    def __normalizeQualifications(self, ranking):
        # find highest qualification of candidate
        qualifications = [ranking[i].qualification for i in range(len(ranking))]
        highest = max(qualifications)
        for candidate in ranking:
            candidate.qualification = candidate.qualification / highest
            candidate.originalQualification = candidate.originalQualification / highest

    def __determineIfProtected(self, r, protAttr):
        """
        takes a JSON profile and finds if the person belongs to the protected group
        Parameter:
        ---------
        r : JSON node
        a person description in JSON, everything below node "profile"
        """

        if 'sex' in r['profile'][0]:
            if r['profile'][0]['sex'] == protAttr:
                return True
            else:
                return False
        else:
            print('>>> undetermined\n')
            return False

    def __determineWorkMonths(self, r):
        """
        takes a person's profile as JSON node and computes the total amount of work months this
        person has
        Parameters:
        ----------
        r : JSON node
        """

        total_working_months = 0  # ..of that profile
        job_duration = 0

        if len(r['profile'][0]) > 4:  # a job is on the profile
            list_of_Jobs = r['profile'][0]['jobs']
            # print('profile summary' + str(r['profile'][0]['jobs']))
            for count in range(0, len(list_of_Jobs)):
                if len(list_of_Jobs[count]) > 3:  # an exact duration is given at 5 nodes!

                    job_duration_string = list_of_Jobs[count]['jobDates']
                    if job_duration_string == 'bis heute':
                        # print('job with no dates found - will be count for ' + str(job_with_no_dates) + ' months.')
                        job_duration = self.EDUCATION_OR_JOB_WITH_NO_DATES

                    else:
                        job_start_string, job_end_string = job_duration_string.split(' - ')

                        if len(job_start_string) == 4:
                            job_start = datetime.datetime.strptime(job_start_string, "%Y")
                        elif len(job_start_string) == 7:
                            job_start = datetime.datetime.strptime(job_start_string, "%m/%Y")
                        else:
                            print("error reading start date")

                        if len(job_end_string) == 4:
                            job_end = datetime.datetime.strptime(job_end_string, "%Y")
                        elif len(job_end_string) == 7:
                            job_end = datetime.datetime.strptime(job_end_string, "%m/%Y")
                        else:
                            print("error reading end date")

                        if job_end - job_start == 0:
                            delta = self.EDUCATION_OR_JOB_WITH_SAME_YEAR
                        else:
                            delta = job_end - job_start

                        job_duration = math.ceil(delta.total_seconds() / 2629743.83)

                total_working_months += job_duration
        else:
            print('-no jobs on profile-')

        return total_working_months

    def __determineEduMonths(self, r):
        """
        takes a person's profile as JSON node and computes the total amount of work months this
        person has
        Parameters:
        ----------
        r : JSON node
        """

        total_education_months = 0  # ..of that profile
        edu_duration = 0

        if 'education' in r:  # education info is on the profile
            list_of_edu = r['education']  # edu child nodes {institution, url, degree, eduDuration}
            # print('education summary' + str(r['education']))
            for count in range(0, len(list_of_edu)):
                if 'eduDuration' in list_of_edu[count]:  # there are education dates

                    edu_duration_string = list_of_edu[count]['eduDuration']
                    if edu_duration_string == ('bis heute' or None or ''):
                        edu_duration = self.EDUCATION_OR_JOB_WITH_NO_DATES
                    else:
                        edu_start_string, edu_end_string = edu_duration_string.split(' - ')

                        if len(edu_start_string) == 4:
                            edu_start = datetime.datetime.strptime(edu_start_string, "%Y")
                        elif len(edu_start_string) == 7:
                            edu_start = datetime.datetime.strptime(edu_start_string, "%m/%Y")
                        else:
                            print("error reading start date")

                        if len(edu_end_string) == 4:
                            edu_end = datetime.datetime.strptime(edu_end_string, "%Y")
                        elif len(edu_end_string) == 7:
                            edu_end = datetime.datetime.strptime(edu_end_string, "%m/%Y")
                        else:
                            print("error reading end date")

                        if edu_end - edu_start == 0:
                            delta = self.EDUCATION_OR_JOB_WITH_SAME_YEAR
                        else:
                            delta = edu_end - edu_start

                        edu_duration = math.ceil(delta.total_seconds() / 2629743.83)

                        # print(job_duration_string)
                        # print('this job: ' + str(job_duration))

                else:
                    edu_duration = self.EDUCATION_OR_JOB_WITH_NO_DATES

                total_education_months += edu_duration
                # print('total jobs: ' + str(total_working_months))

            # print("studying: " + str(total_education_months))
        else:
            print('-no education on profile-')

        return total_education_months



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

from __future__ import annotations

import numpy as np
import numpy.typing as npt
from typing import Any

from sklearn.utils.validation import check_is_fitted
from sklearn.utils._param_validation import HasMethods
from sklearn.metrics.pairwise import rbf_kernel
from sklearn.svm import SVC

from ._base import FairEstimator


def linear_kernel(x1, x2):
    return np.dot(x1, np.transpose(x2))


class LinearFairERM(FairEstimator):
    """
    Empirical risk minimization under fairness constraints

    Reference:
        https://proceedings.neurips.cc/paper/2018/file/83cdcec08fbf90370fcf53bdd56604ff-Paper.pdf
    Code adopted from:
        https://github.com/jmikko/fair_ERM

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        A list of class labels known to the classifier.

    s_classes_ : ndarray of shape (n_sensitive_group,)
        A list of sensitive classes known to LabelBias during training.

    fitted_estimator_ : an object of the classifier
        Fitted estimator.

    n_features_in_ : int
        Number of features seen during `fit`.

    feature_names_in_ : ndarray of shape (n_features_in_,)
        Names of features seen during `fit`. Defined only when X has feature names that are all strings.

    Examples
    --------
    >>> from sklearn.preprocessing import StandardScaler
    >>> from fairpy.dataset import Adult
    >>> from fairpy.model import LinearFairERM
    >>> dataset = Adult()
    >>> split_data = dataset.split()
    >>> model = LinearFairERM()
    >>> model.fit(split_data.X_train, split_data.y_train, split_data.s_train)
    >>> model.predict(split_data.X_test)
    """

    _parameter_constraints = {
        "estimator": [HasMethods(["fit"]), None],
    }

    def __init__(self, estimator: Any = None):
        """
        Parameters
        ----------
        estimator : a callable object with some requirements, default=None
            Estimator is the base classifier, and LinearFairERM is a wrapper function over this classifier.
            The estimator should implement 'fit' method (and more if needed) for
            model's training. 'fit' should accept arguments: X, y. If not specific,
            use Support Vector Machine as the estimator.

        """

        self.estimator = estimator

    def _feat_trans(self, X: npt.NDArray) -> npt.NDArray:
        """ Linear feature transformations """

        trans_X = X - np.outer((X[:, self._max_idx] / self._u[self._max_idx]), self._u)
        trans_X = np.delete(trans_X, self._max_idx, 1)

        return trans_X

    def fit(self, X: npt.ArrayLike, y: npt.ArrayLike, s: npt.ArrayLike) -> LinearFairERM:
        """
        Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training vector, where `n_samples` is the number of samples and `n_features` is the number of features.

        y : array-like of shape (n_samples,)
            Target vector relative to X.

        s : array-like of shape (n_samples,)
            Sensitive attributes relative to X.

        Returns
        -------
        self
            Fitted estimator.
        """

        self._validate_params()
        X, y, s = self._validate_data(X, y, s)
        y = self._validate_cls_y(y)
        s = self._validate_grp_s(s)

        if self.estimator is None:
            model = SVC(kernel="linear")
        else:
            model = self.estimator

        idx_grp_0 = np.where(s == 0)
        idx_grp_1 = np.where(s == 1)
        idx_y_pos = np.where(y == 1)
        idx_y_pos_grp_0 = np.intersect1d(idx_grp_0, idx_y_pos)
        idx_y_pos_grp_1 = np.intersect1d(idx_grp_1, idx_y_pos)

        feat_grp_0 = np.mean(X[idx_y_pos_grp_0], axis=0)
        feat_grp_1 = np.mean(X[idx_y_pos_grp_1], axis=0)

        self._u = feat_grp_1 - feat_grp_0
        self._max_idx = np.argmax(self._u)

        trans_X = self.feat_trans(X)
        model.fit(trans_X, y)

        self.fitted_estimator_ = model

        return self

    def feat_trans(self, X: npt.ArrayLike) -> npt.NDArray:
        """
        Feature transformation for fairness.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data matrix for which we want to get the transformations.

        Returns
        -------
        trans_X : ndarray of shape (n_samples,)
            Vector containing the class labels for each sample.
        """

        check_is_fitted(self)
        X = self._validate_data(X, reset=False)

        return self._feat_trans(X)

    def predict(self, X) -> Any:
        """
        Predict class labels for samples in X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data matrix for which we want to get the predictions.

        Returns
        -------
        y_pred : ndarray of shape (n_samples,)
            Vector containing the class labels for each sample.
        """

        if not hasattr(self.estimator, "predict"):
            raise ValueError("Estimator is expected to have corresponding method 'predict' when calling 'predict'.")
        trans_X = self.feat_trans(X)

        return self.estimator.predict(trans_X)

    def predict_proba(self, X) -> Any:
        """
        Probability estimates.

        The returned estimates for all classes are ordered by the
        label of classes.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Vector to be scored, where `n_samples` is the number of samples and
            `n_features` is the number of features.

        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Returns the probability of the sample for each class in the model,
            where classes are ordered as they are in ``self.classes_``.
        """

        if not hasattr(self.estimator, "predict_proba"):
            raise ValueError(
                "Estimator is expected to have corresponding method 'predict_proba' when calling 'predict_proba'.")
        trans_X = self.feat_trans(X)

        return self.estimator.predict_proba(trans_X)

    def predict_log_proba(self, X) -> Any:
        """
        Predict logarithm of probability estimates.

        The returned estimates for all classes are ordered by the
        label of classes.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Vector to be scored, where `n_samples` is the number of samples and
            `n_features` is the number of features.

        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Returns the logarithm of probability of the sample for each class in the model,
            where classes are ordered as they are in ``self.classes_``.
        """

        if not hasattr(self.estimator, "predict_log_proba"):
            raise ValueError(
                "Estimator is expected to have corresponding method 'predict_log_proba' when calling 'predict_log_proba'.")
        trans_X = self.feat_trans(X)

        return self.estimator.predict_log_proba(trans_X)

    def decision_function(self, X) -> Any:
        """
        Predict confidence scores for samples.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data matrix for which we want to get the confidence scores.

        Returns
        -------
        scores : ndarray of shape (n_samples,)
            Confidence scores per `(n_samples, n_classes)` combination. In the
            binary case, confidence score for `self.classes_[1]` where >0 means
            this class would be predicted.
        """

        if not hasattr(self.estimator, "decision_function"):
            raise ValueError(
                "Estimator is expected to have corresponding method 'decision_function' when calling 'decision_function'.")
        trans_X = self.feat_trans(X)

        return self.estimator.decision_function(trans_X)

    def _more_tags(self):
        return {
            'binary_only': True,
            'requires_y': True,
            'requires_s': True,
            "y_types": ["binary"],
            "s_types": ["binary"],
        }



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import os
import sys
import pandas as pd
from typing import Sequence, Union

from ._base import TabDataset


class Oulad(TabDataset):
    """ https://analyse.kmi.open.ac.uk/open_dataset """

    file2url = {
        "anonymisedData.zip": "https://analyse.kmi.open.ac.uk/open_dataset/download",
    }

    categorical_feat = [
        "code_module",
        "code_presentation",
        "gender",
        "region",
        "highest_education",
        "imd_band",
        "age_band",
        "disability",
    ]

    numerical_feat = [
        "num_of_prev_attempts",
        "studied_credits",
    ]

    drop_feat = ["id_student"]
    na_values = (" ")

    label_name = "final_result"
    label_mapping = ("Fail", "Pass")

    dataset_name = "oulad"

    def __init__(
            self,
            sen_feat: Union[str, Sequence[str]] = "gender",
            dir_path: str = os.path.join(sys.path[1], "fairpy/data/oulad"),
            download: bool = False,
    ):
        self._check_and_download_data(dir_path, self.file2url, download)
        df = pd.read_csv(os.path.join(dir_path, "studentInfo.csv"))

        super(Oulad, self).__init__(
            df=df,
            label_name=self.label_name,
            sen_feat=sen_feat,
            cat_feat=self.categorical_feat,
            num_feat=self.numerical_feat,
            dataset_name=self.dataset_name,
            custom_preprocessing=self.default_preprocessing,
            na_values=self.na_values,
            drop_feat=self.drop_feat,
            label_mapping=self.label_mapping,
        )

    @staticmethod
    def default_preprocessing(df: pd.DataFrame) -> pd.DataFrame:
        def final_result(row):
            return 'Pass' if row['final_result'] == "Distinction" else row['final_result']

        df['final_result'] = df.apply(lambda row: final_result(row), axis=1)
        df = df[(df["final_result"] == "Pass") | (df["final_result"] == "Fail")]
        df.reset_index(inplace=True, drop=True)

        return df



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import scipy
from scipy.spatial.distance import pdist, cdist, squareform
from numbers import Real, Integral
from typing import Union, Sequence

from sklearn.utils.validation import check_is_fitted
from scipy.optimize import minimize
from sklearn.utils._param_validation import Interval

from ._base import FairEstimator


class IFair(FairEstimator):
    """
    iFair: Learning Individually Fair Data Representations for Algorithmic Decision Making

    Reference:
        https://arxiv.org/pdf/1806.01059.pdf
    Code adopted from:
        https://github.com/plahoti-lgtm/iFair

    The time complexity is O(N^2) for every optimization iteration.

    Attributes
    ----------
    n_features_in_ : int
        Number of features seen during `fit`.

    feature_names_in_ : ndarray of shape (n_features_in_,)
        Names of features seen during `fit`. Defined only when X has feature names that are all strings.

    s_idx_ : list of int
        The index(es) of the sensitive attribute(s) in the data matrix.

    opt_params_ : ndarray of shape (n_features * n_centroids + n_features)
        Solved coefficient in probabilistic clustering

    Examples
    --------
    >>> from sklearn.preprocessing import StandardScaler
    >>> from fairpy.dataset import Adult
    >>> from fairpy.model import LabelBias
    >>> dataset = Adult()
    >>> split_data = dataset.split()
    >>> model = IFair()
    >>> model.fit(split_data.X_train)
    >>> fair_data = model.transform(split_data.X_train)
    """

    _parameter_constraints = {
        "s_idx": [None, Interval(Integral, 0, None, closed="left"), "array-like"],
        "K": [Interval(Integral, 0, None, closed="left")],
        "max_iter": [Interval(Integral, 0, None, closed="left")],
        "restarts": [Interval(Integral, 0, None, closed="left")],
        "epsilon": [Interval(Real, 0, None, closed="left")],
        "w_recon": [Interval(Real, 0, None, closed="left")],
        "w_fair": [Interval(Real, 0, None, closed="left")],
    }

    def __init__(
            self,
            s_idx: Union[int, Sequence[int]] = None,
            K: int = 2,
            max_iter: int = 200,
            restarts: int = 3,
            epsilon: float = 1e-4,
            w_recon: float = 1.,
            w_fair: float = 1.,
    ):
        """
        Parameters
        ----------
        s_idx : int or a list of int, default=None
            The index(es) of the sensitive attribute(s) in the data matrix.

        K : int, default=2
            The number of centroids in probabilistic clustering.

        max_iter : int, default=200
            Maximum numbers of iterations taken for the solvers to converge.

        restarts : int, default=3
            Times to restart the optimization, the final optimal parameters are taken from these restarts.

        epsilon : float, default=1e-4
            Initial weights associated with sensitive attributes when computing pair-wise distances.

        w_recon : float, default=1.
            The weight of data reconstruction loss in the objective function.

        w_fair : float, default=1.
            The weight of individual fairness loss in the objective function.
        """

        self.s_idx = s_idx
        self.K = K
        self.max_iter = max_iter
        self.restarts = restarts
        self.epsilon = epsilon
        self.w_recon = w_recon
        self.w_fair = w_fair

    def _forward(self, params: npt.NDArray, X: npt.NDArray) -> npt.NDArray:
        """ Forward pass """

        alpha = params[:self._M]
        centroids = np.asarray(params[self._M:], dtype=np.float64).reshape((self.K, self._M))

        dist_mat = cdist(X, centroids, "euclidean", w=alpha)
        U = scipy.special.softmax(-dist_mat, axis=1)
        X_hat = U @ centroids

        return X_hat

    def _loss(self, params: npt.NDArray, X: npt.NDArray, D: npt.NDArray) -> float:
        """ Reconstruction loss and individual fairness loss """

        X_hat = self._forward(params, X)
        loss_recon = np.linalg.norm(X - X_hat)

        D_hat = squareform(pdist(X_hat, "euclidean"))
        loss_fair = np.linalg.norm(D - D_hat)

        loss = self.w_recon * loss_recon + self.w_fair * loss_fair

        return loss

    def fit(self, X: npt.ArrayLike) -> IFair:
        """
        Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training vector, where `n_samples` is the number of samples and `n_features` is the number of features.

        Returns
        -------
        self
            Fitted estimator.
        """

        self._validate_params()
        X = self._validate_data(X)

        self._N, self._M = X.shape

        if self.s_idx is not None:
            self.s_idx_ = self._validate_idx(self.s_idx, self._M)
            non_s_idx = list(set([i for i in range(self._M)]) - set(self.s_idx_))
        else:
            self.s_idx_ = None
            non_s_idx = [i for i in range(self._M)]

        # Distance matrix without sensitive attributes
        X_non_s = X[:, non_s_idx]
        D = np.asarray(squareform(pdist(X_non_s, "euclidean")), dtype=np.float64)

        min_obj = None
        opt_params = None
        params_size = self._M + self._M * self.K
        bnd = [(0, None) if i < self._M else (None, None) for i in range(params_size)]

        for _ in range(self.restarts):
            init = np.random.uniform(size=params_size).astype(np.float64)
            if self.s_idx_ is not None:
                init[self.s_idx_] = self.epsilon

            opt = minimize(
                self._loss,
                init,
                args=(X, D),
                method="L-BFGS-B",
                jac=False,
                bounds=bnd,
                options={
                    'maxiter': self.max_iter,
                    'maxfun': self.max_iter,
                    'eps': 1e-3,
                }
            )

            if (min_obj is None) or (opt.fun < min_obj):
                min_obj = opt.fun
                opt_params = opt.x

        self.opt_params_ = opt_params

        return self

    def transform(self, X: npt.ArrayLike) -> npt.NDArray:
        """
        Transform data matrix to fair data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training vector, where `n_samples` is the number of samples and `n_features` is the number of features.

        Returns
        -------
        X : ndarray of shape (n_samples, n_features)
            Debiased training vector.
        """

        check_is_fitted(self)
        X = self._validate_data(X, reset=False)

        return self._forward(self.opt_params_, X)

    def fit_transform(self, X: npt.ArrayLike) -> npt.NDArray:
        """
        Fit the estimator and transform data matrix to fair data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training vector, where `n_samples` is the number of samples and `n_features` is the number of features.

        Returns
        -------
        X : ndarray of shape (n_samples, n_features)
            Debiased training vector.
        """

        self.fit(X)
        return self._forward(self.opt_params_, X)



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import math
import copy
import random
import numpy as np
import numpy.typing as npt
from copy import deepcopy
from numbers import Real
import networkx as nx
from collections import OrderedDict, Counter
from typing import Sequence, Union, Any, Optional, Dict, NewType, Tuple, Callable, Mapping, List

from sklearn.utils._param_validation import Interval

from ._base import FairEstimator

SenFeat = NewType('SenFeat', Union[int, float, str, Sequence])  # sensitive feature
ColFeat = NewType('ColFeat', Union[int, float, str])  # feature at current column


def get_median(input: Sequence) -> Any:
    sorted_input = sorted(deepcopy(input))

    if len(sorted_input) % 2 == 0:
        return sorted_input[len(sorted_input) // 2 - 1]
    else:
        return sorted_input[len(sorted_input) // 2]


class CategoricalFeat():
    """ Modeling a categorical feature """

    def __init__(self, data: npt.NDArray):
        self.data = data
        self.bin_data = dict(Counter(self.data))
        self.val2idx = {val: np.where(self.data == val)[0].tolist() for val in self.bin_data.keys()}
        self.N_bin = len(self.bin_data)

    def create_graph(self, generator: Callable) -> nx.classes.digraph.DiGraph:
        dg = nx.DiGraph()
        dg.add_node("s")
        dg.add_node("t")

        for i, (col_val, cnt) in enumerate(self.bin_data.items()):
            dg.add_node(i)
            dg.add_node(i + self.N_bin)
            dg.add_edge("s", i, capacity=cnt, weight=0)
            dg.add_edge(i + self.N_bin, "t", capacity=generator(col_val), weight=0)

        dg.add_node(2 * self.N_bin)
        dg.add_edge(2 * self.N_bin, "t", capacity=len(self), weight=0)

        for i in range(self.N_bin):
            for j in range(self.N_bin, 2 * self.N_bin):
                if i + self.N_bin == j:
                    dg.add_edge(i, j, weight=0)
                else:
                    dg.add_edge(i, j, weight=1)

            dg.add_edge(i, 2 * self.N_bin, weight=2)

        self.dg = dg

        return dg

    def repair(self) -> Tuple[npt.NDArray, int]:
        min_cost_flow = nx.max_flow_min_cost(self.dg, "s", "t")

        val2idx_ = copy.deepcopy(self.val2idx)
        repair_data = np.asarray(["PLACEHOLDER_DIREMOVER_FAIRPY" for _ in range(len(self))])
        repair_bin_dict = {}
        candidates = list(self.bin_data.keys())
        overflow = 0

        for i, (col_val, cnt) in enumerate(self.bin_data.items()):
            overflow += min_cost_flow[i][2 * self.N_bin]

            for j in range(self.N_bin, 2 * self.N_bin):
                edgeflow = min_cost_flow[i][j]
                group = random.sample(val2idx_[col_val], int(edgeflow))

                q = j - self.N_bin

                for e in group:
                    val2idx_[col_val].remove(e)
                    repair_data[e] = candidates[q]

                if q in repair_bin_dict:
                    repair_bin_dict[q].extend(group)
                else:
                    repair_bin_dict[q] = group

        return repair_data, overflow

    def __len__(self):
        return len(self.data)


class DIRemover(FairEstimator):
    """
    Certifying and Removing Disparate Impact

    Reference:
        https://dl.acm.org/doi/pdf/10.1145/2783258.2783311
    Code adopted from:
        https://github.com/algofairness/BlackBoxAuditing

    # TODO: bugs on data types in numpy

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        A list of class labels known to the classifier.

    s_classes_ : ndarray of shape (n_sensitive_group,)
        A list of sensitive classes known to LabelBias during training.

    n_features_in_ : int
        Number of features seen during `fit`.

    feature_names_in_ : ndarray of shape (n_features_in_,)
        Names of features seen during `fit`. Defined only when X has feature names that are all strings.

    s_idx_ : list of int
        The index(es) of the sensitive attribute(s) in the data matrix used to repair data.

    repair_feat_idx_ : list of int
        The index(es) of the feature(s) to be repaired in the data matrix.

    num_feat_idx_ : list of int
        The index(es) of the numerical feature(s) in the data matrix.

    Examples
    --------
    >>> from fairpy.dataset import Adult
    >>> from fairpy.model import DIRemover
    >>> dataset = Adult()
    >>> split_data = dataset.split()
    >>> model = DIRemover(s_idx=dataset.feat_idx.sen_idx, num_feat_idx=0)
    >>> new_X_train = model.fit_transform(split_data.X_train)
    """

    _parameter_constraints = {
        "s_idx": [None, Interval(Real, 0, None, closed="left"), "array-like"],
        "repair_feat_idx": [None, Interval(Real, 0, None, closed="left"), "array-like"],
        "num_feat_idx": [None, Interval(Real, 0, None, closed="left"), "array-like"],
        "repair_level": [Interval(Real, 0, 1, closed="both")],
    }

    def __init__(
            self,
            s_idx: Union[int, Sequence[int]],
            repair_feat_idx: Optional[Union[int, Sequence[int]]] = None,
            num_feat_idx: Optional[Union[int, Sequence[int]]] = None,
            repair_level: float = 1.,
    ):
        """
        Parameters
        ----------
        s_idx : int or a list of int
            The index(es) of the sensitive attribute(s) in the data matrix used to repair data.

        repair_feat_idx: int or a list of int
            The index(es) of the feature(s) to be repaired in the data matrix.

        num_feat_idx : int or a list of int
            The index(es) of the numerical feature(s) in the data matrix.

        repair_level : float
            The strength of repair to features, range from 0 to 1 and 1 indicates most repair.
        """

        self.s_idx = s_idx
        self.repair_feat_idx = repair_feat_idx
        self.num_feat_idx = num_feat_idx
        self.repair_level = repair_level

    def _validate_all_idx(self, X: npt.NDArray) -> None:
        """ Validate indexes in s_idx, repair_feat_idx, and num_feat_idx """

        self.s_idx_ = self._validate_idx(self.s_idx, X.shape[1])

        if self.repair_feat_idx is None:
            self.repair_feat_idx_ = (i for i in range(X.shape[1]))
        else:
            self.repair_feat_idx_ = self._validate_idx(self.repair_feat_idx, X.shape[1])

        if self.num_feat_idx is not None:
            self.num_feat_idx_ = self._validate_idx(self.num_feat_idx, X.shape[1])
        else:
            self.num_feat_idx_ = None

        return

    def _repair_num_feat(
            self,
            x: npt.NDArray,
            unique_s: Sequence[SenFeat],
            s2val2idx: Mapping[SenFeat, Mapping[ColFeat, Sequence[int]]],
    ) -> npt.NDArray:
        """ Repair a numerical feature x """

        # minimal number of unique column values for all sensitive attributes
        N_quantiles = min(len(set_) for set_ in s2val2idx.values())
        quantile_unit = 1. / N_quantiles

        s2offsets = {s: 0 for s in unique_s}

        ranked_col_val = sorted(np.unique(x))
        for quantile in range(N_quantiles):
            median_at_curr_quantile = []
            s2idx_curr_quantile = {}

            for s in unique_s:
                N_col_vals = len(s2val2idx[s])
                offset = int(round(s2offsets[s] * N_col_vals))
                number_to_get = int(round((s2offsets[s] + quantile_unit) * N_col_vals) - offset)
                s2offsets[s] += quantile_unit

                if number_to_get > 0:
                    # select column values at current quantile
                    select_col_val = list(s2val2idx[s].keys())[offset:offset + number_to_get]
                    s2idx_curr_quantile[s] = [idx for val in select_col_val for idx in s2val2idx[s][val]]
                    median_at_curr_quantile.append(get_median(select_col_val))

            median = get_median(median_at_curr_quantile)
            median_rank = ranked_col_val.index(median)

            for s in unique_s:
                for row_idx in s2idx_curr_quantile[s]:
                    # repair column values at current quantile
                    curr_val = x[row_idx]
                    curr_val_rank = ranked_col_val.index(curr_val)
                    distance = median_rank - curr_val_rank
                    distance_to_repair = int(round(distance * self.repair_level))
                    repaired_val = ranked_col_val[curr_val_rank + distance_to_repair]
                    x[row_idx] = repaired_val

        return x

    def _gen_desired_dist(
            self,
            s: SenFeat,
            col_val: ColFeat,
            cnt_norm: Mapping[ColFeat, Mapping[SenFeat, float]],
            median: Mapping[ColFeat, float],
            mode: ColFeat,
            is_sen_feat: bool,
    ) -> float:
        if is_sen_feat:
            return 1 if col_val == mode else (1 - self.repair_level) * cnt_norm[col_val][s]
        else:
            return (1 - self.repair_level) * cnt_norm[col_val].get(s, 0) + (self.repair_level * median[col_val])

    def _gen_desired_cnt(
            self,
            s: SenFeat,
            col_val: ColFeat,
            s2cat_feat: Mapping[SenFeat, CategoricalFeat],
            cnt_dict: Mapping[ColFeat, Mapping[SenFeat, int]],
            median: Mapping[ColFeat, float],
    ) -> int:
        N = len(s2cat_feat[s])
        cnt = cnt_dict[col_val].get(s, 0)
        des_cnt = math.floor(((1 - self.repair_level) * cnt) + (self.repair_level) * median[col_val] * N)

        return des_cnt

    def _flow(
            self,
            unique_s: Sequence[SenFeat],
            s2cat_feat: Mapping[SenFeat, CategoricalFeat],
            repair_generator: Callable,
    ) -> Mapping[SenFeat, Tuple[npt.NDArray, int]]:

        s2flow = {}
        for i, s in enumerate(unique_s):
            cat_feat = s2cat_feat[s]
            generator = lambda col_val: repair_generator(s, col_val)
            cat_feat.create_graph(generator)
            new_feat, overflow = cat_feat.repair()
            s2flow[s] = (new_feat, overflow)

        return s2flow

    def _assign_overflow(
            self,
            unique_s: Sequence[SenFeat],
            unique_col_val: Sequence[ColFeat],
            s2flow: Mapping[SenFeat, Tuple[np.ndarray, int]],
            repair_generator: Callable,
    ) -> Mapping[SenFeat, npt.NDArray]:

        assigned_overflow = {}
        s2repaired = {s: flow[0] for s, flow in s2flow.items()}

        for s in unique_s:
            dist_generator = lambda col_val: repair_generator(s, col_val)
            col_val_props = list(map(dist_generator, unique_col_val))

            if all(e == 0 for e in col_val_props):
                col_val_props = [1. / len(col_val_props)] * len(col_val_props)

            col_val_props = [e / float(sum(col_val_props)) for e in col_val_props]

            assigned_overflow[s] = {}
            for i in range(int(s2flow[s][1])):
                N = random.uniform(0, 1)
                cat_index = 0
                tally = 0

                for j in range(len(col_val_props)):
                    val = col_val_props[j]
                    if N < (tally + val):
                        cat_index = j
                        break
                    tally += val

                assigned_overflow[s][i] = unique_col_val[cat_index]

            cnt = 0
            for i, val in enumerate(s2flow[s][0]):
                if val == "PLACEHOLDER_DIREMOVER_FAIRPY":
                    s2repaired[s][i] = assigned_overflow[s][cnt]
                    cnt += 1

        return s2repaired

    def _repair_cat_feat(
            self,
            x: npt.NDArray,
            unique_s: Sequence[SenFeat],
            s2row: Mapping[SenFeat, Sequence[int]],
            s2val2idx: Mapping[SenFeat, Mapping[ColFeat, List[int]]],
    ) -> npt.NDArray:
        """ Repair a categorical feature """

        unique_col_val = np.unique(x)
        mode = max(s2row, key=lambda k: len(s2row[k]))

        # Construct CatFeat for each sensitive attribute
        s2cat_feat = {}
        for s, row_idx in s2row.items():
            val = x[row_idx]
            s2cat_feat[s] = CategoricalFeat(val)

        # Calculate and normalize the joint distribution over column value and sensitive attribute
        cnt_dict: Dict[ColFeat, OrderedDict[SenFeat, int]] = {}
        cnt_norm_dict: Dict[ColFeat, OrderedDict[SenFeat, float]] = {}
        for col_val in unique_col_val:
            temp_cnt = {}
            temp_cnt_norm = {}
            for s in unique_s:
                if col_val in s2val2idx[s]:
                    temp_cnt[s] = len(s2val2idx[s][col_val])
                    temp_cnt_norm[s] = len(s2val2idx[s][col_val]) / len(s2row[s])

            cnt_dict[col_val] = OrderedDict(data=sorted(temp_cnt.items(), key=lambda x: x[1]))
            cnt_norm_dict[col_val] = OrderedDict(sorted(temp_cnt_norm.items(), key=lambda x: x[1]))

        # TODO: possible computational error in median
        def list_pad(col_val):
            temp = list(cnt_norm_dict[col_val].values())
            if len(temp) == 1:
                temp.insert(0, 0)
            return temp

        median = {col_val: get_median(list_pad(col_val)) for col_val in unique_col_val}
        # median = {'A': 0.3333333333333333, 'B': 0.30000000000000004, 'C': 0}
        # print("median", median)

        dist_generator = lambda s, col_val: self._gen_desired_dist(s, col_val, cnt_norm_dict, median, mode, False)
        cnt_generator = lambda s, col_val: self._gen_desired_cnt(s, col_val, s2cat_feat, cnt_dict, median)

        s2flow = self._flow(
            unique_s=unique_s,
            s2cat_feat=s2cat_feat,
            repair_generator=cnt_generator,
        )

        s2repaired = self._assign_overflow(
            unique_s=unique_s,
            unique_col_val=unique_col_val,
            s2flow=s2flow,
            repair_generator=dist_generator,
        )

        for s in unique_s:
            for i, idx in enumerate(s2row[s]):
                x[idx] = s2repaired[s][i]

        return x

    def fit_transform(self, X: npt.ArrayLike) -> npt.NDArray:
        """
        Fit DIRemover and return repaired data matrix.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training vector, where `n_samples` is the number of samples and `n_features` is the number of features.

        Returns
        -------
        X : array-like of shape (n_samples, n_features)
            Repaired data matrix.
        """

        self._validate_params()
        X = self._validate_data(X)
        self._validate_all_idx(X)

        X = np.copy(X)

        s_cols = X[:, self.s_idx_].astype('U')
        unique_s = np.unique(s_cols, axis=0)
        unique_s = tuple(tuple(s) for s in unique_s)

        # sensitive attribute -> indexes of rows containing the sensitive attribute
        s2row: Dict[SenFeat, Sequence[int]] = {
            s: np.where(np.prod(X[:, self.s_idx_] == s, axis=1) == 1)[0] for s in unique_s
        }

        # repair data on a column basis
        for col_idx in self.repair_feat_idx_:
            s2val2idx = {}  # sensitive attribute -> unique values in the current column -> corresponding row indexes
            for s in unique_s:
                unique_val = np.unique(X[s2row[s], col_idx])
                curr_dict = {
                    val: np.intersect1d(np.where(X[:, self.s_idx_] == s)[0], np.where(X[:, col_idx] == val)[0]).tolist()
                    for val in unique_val
                }
                s2val2idx[s] = curr_dict

            # sort column values in a ascending order for numerical features
            s2val2idx = {s: OrderedDict(sorted(dict_.items())) for s, dict_ in s2val2idx.items()}

            if self.num_feat_idx_ is not None and col_idx in self.num_feat_idx_:
                X[:, col_idx] = self._repair_num_feat(X[:, col_idx], unique_s, s2val2idx)
            else:
                X[:, col_idx] = self._repair_cat_feat(X[:, col_idx], unique_s, s2row, s2val2idx)

        return X



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import abc
import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy.stats import binom
from numbers import Real, Integral
from typing import List

from sklearn.utils._param_validation import Interval

from ._base import FairEstimator


def compute_aux_mtable(mtable: pd.DataFrame) -> pd.DataFrame:
    """
    Stores the inverse of an mTable entry and the size of the block with respect to the inverse
    """

    if not (isinstance(mtable, pd.DataFrame)):
        raise TypeError("Internal mtable must be a DataFrame")

    aux_mtable = pd.DataFrame(columns=["inv", "block"])
    last_m_seen = 0
    last_position = 0
    for position in range(1, len(mtable)):
        if position % 2000 == 0:
            print("Computing m inverse: {:.0f} of {:.0f}".format(position, len(mtable)))
        if mtable.at[position, "m"] == last_m_seen + 1:
            last_m_seen += 1
            aux_mtable.loc[position] = [position, position - last_position]
            last_position = position
        elif mtable.at[position, "m"] != last_m_seen:
            raise RuntimeError("Inconsistent mtable")

    return aux_mtable


class MTableFailProbPair:
    """
    Encapsulation of all parameters for the interim mtables
    """

    def __init__(self, k, p, alpha, fail_prob, mtable):
        self.k = k
        self.p = p
        self.alpha = alpha
        self.fail_prob = fail_prob
        self.mtable = mtable

    def mass_of_mtable(self):
        return self.mtable['m'].sum()


class LegalAssignmentKey:
    """
    Utility class for the recursive fail prob
    """

    def __init__(self, remaining_candidates, remaining_block_sizes, current_block_number, candidates_assigned_so_far):
        self.remaining_candidates = remaining_candidates
        self.remaining_block_sizes = remaining_block_sizes
        self.current_block_number = current_block_number
        self.candidates_assigned_so_far = candidates_assigned_so_far

    def __eq__(self, other):
        if self.remaining_candidates != other.remaining_candidates:
            return False
        if self.current_block_number != other.current_block_number:
            return False
        if self.candidates_assigned_so_far != other.candidates_assigned_so_far:
            return False
        if self.remaining_block_sizes != other.remaining_block_sizes:
            return False
        return True

    def __hash__(self):
        return int(self.remaining_candidates + len(
            self.remaining_block_sizes) << 16) + self.current_block_number + self.candidates_assigned_so_far


class FailProbabilityCalculator(abc.ABC):
    """
    Base class for the fail probability calculation
    """

    def __init__(self, k, p, alpha):
        self.k = k
        self.p = p
        self.alpha = alpha

        self.pmf_cache = {}

    @abc.abstractmethod
    def calculate_fail_probability(self, mtable):
        raise NotImplementedError("This is an abstract method. Implement it!")

    def get_from_pmf_cache(self, trials, successes):
        key = (trials, successes)
        if not key in self.pmf_cache:
            # TODO: Check the documentation if this is fine
            self.pmf_cache[key] = binom.pmf(k=successes, n=trials, p=self.p)
        return self.pmf_cache[key]


class RecursiveNumericFailProbabilityCalculator(FailProbabilityCalculator):
    """
    Recursive calculation of fail probability
    """

    EPS = 1e-12

    def __init__(self, k, p, alpha):
        super().__init__(k, p, alpha)

        self.legal_assignment_cache = {}

    def adjust_alpha(self):
        a_min = 0
        a_max = self.alpha
        a_mid = (a_min + a_max) / 2

        minb = self._compute_boundary(a_min)
        maxb = self._compute_boundary(a_max)
        midb = self._compute_boundary(a_mid)

        while minb.mass_of_mtable() < maxb.mass_of_mtable() and midb.fail_prob != self.alpha:
            if midb.fail_prob < self.alpha:
                a_min = a_mid
                minb = self._compute_boundary(a_min)
            elif midb.fail_prob > self.alpha:
                a_max = a_mid
                maxb = self._compute_boundary(a_max)

            a_mid = (a_min + a_max) / 2
            midb = self._compute_boundary(a_mid)

            max_mass = maxb.mass_of_mtable()
            min_mass = minb.mass_of_mtable()
            mid_mass = midb.mass_of_mtable()

            if max_mass - min_mass == 1 or maxb.alpha - minb.alpha <= RecursiveNumericFailProbabilityCalculator.EPS:
                min_diff = abs(minb.fail_prob - self.alpha)
                max_diff = abs(maxb.fail_prob - self.alpha)

                if min_diff <= max_diff:
                    return minb
                else:
                    return maxb

            if max_mass - mid_mass == 1 and mid_mass - min_mass == 1:
                min_diff = abs(minb.fail_prob - self.alpha)
                max_diff = abs(maxb.fail_prob - self.alpha)
                mid_diff = abs(midb.fail_prob - self.alpha)

                if mid_diff <= max_diff and mid_diff <= min_diff:
                    return midb
                if min_diff <= mid_diff and min_diff <= max_diff:
                    return minb
                else:
                    return maxb

        return midb

    def calculate_fail_probability(self, mtable):
        """
        Analytically calculates the fail probability of the mtable
        """
        aux_mtable = compute_aux_mtable(mtable)
        max_protected = aux_mtable['block'].sum()
        block_sizes = aux_mtable['block'].tolist()  # [1:]
        success_prob = self._find_legal_assignments(max_protected, block_sizes)
        return 0 if success_prob == 0 else (1 - success_prob)

    def _compute_boundary(self, alpha):
        """
        Returns a tuple of (k, p, alpha, fail_prob, mtable)
        """
        mtable = MTableGenerator(self.k, self.p, alpha, False).mtable_as_dataframe()
        fail_prob = self.calculate_fail_probability(mtable)
        return MTableFailProbPair(self.k, self.p, alpha, fail_prob, mtable)

    def _find_legal_assignments(self, number_of_candidates, block_sizes):
        return self._find_legal_assignments_aux(number_of_candidates, block_sizes, 1, 0)

    def _find_legal_assignments_aux(
            self,
            number_of_candidates,
            block_sizes,
            current_block_number,
            candidates_assigned_so_far,
    ):
        if len(block_sizes) == 0:
            return 1

        min_needed_this_block = current_block_number - candidates_assigned_so_far
        if min_needed_this_block < 0:
            min_needed_this_block = 0

        max_possible_this_block = min(block_sizes[0], number_of_candidates)

        assignments = 0
        new_remaining_block_sizes = block_sizes[1:]
        for items_this_block in range(min_needed_this_block, max_possible_this_block + 1):
            new_remaining_candidates = number_of_candidates - items_this_block

            suffixes = self._calculate_legal_assignments_aux(
                new_remaining_candidates,
                new_remaining_block_sizes,
                current_block_number + 1,
                candidates_assigned_so_far + items_this_block,
            )

            assignments += self.get_from_pmf_cache(max_possible_this_block, items_this_block) * suffixes

        return assignments

    def _calculate_legal_assignments_aux(
            self,
            remaining_candidates,
            remaining_block_sizes,
            current_block_number,
            candidates_assigned_so_far,
    ):
        key = LegalAssignmentKey(
            remaining_candidates,
            remaining_block_sizes,
            current_block_number,
            candidates_assigned_so_far,
        )

        if not key in self.legal_assignment_cache:
            self.legal_assignment_cache[key] = self._find_legal_assignments_aux(
                remaining_candidates,
                remaining_block_sizes,
                current_block_number,
                candidates_assigned_so_far,
            )

        return self.legal_assignment_cache[key]


class MTableGenerator():
    def __init__(self, K, P, alpha, adjust_alpha):
        self.k = K
        self.p = P
        self.alpha = alpha
        self.adjust_alpha = adjust_alpha

        if self.adjust_alpha:
            fail_prob_pair = RecursiveNumericFailProbabilityCalculator(K, P, alpha).adjust_alpha()
            self.adjusted_alpha = fail_prob_pair.alpha
            self._mtable = fail_prob_pair.mtable
        else:
            self.adjusted_alpha = alpha
            self._mtable = self._compute_mtable()

    def mtable_as_list(self):
        return [int(i) for i in self._mtable.m.tolist()]

    def mtable_as_dataframe(self):
        return self._mtable

    def m(self, k):
        if k < 1:
            raise ValueError("Parameter k must be at least 1")
        elif k > self.k:
            raise ValueError("Parameter k must be at most {0}".format(self.k))

        result = binom.ppf(self.adjusted_alpha if self.adjust_alpha else self.alpha, k, self.p)
        return 0 if result < 0 else result

    def _compute_mtable(self):
        """ Computes a table containing the minimum number of protected elements
            required at each position
        """
        mtable = pd.DataFrame(columns=["m"])
        for i in range(1, self.k + 1):
            if i % 2000 == 0:
                print("Computing m: {:.0f} of {:.0f}".format(i, self.k))
            mtable.loc[i] = [self.m(i)]
        return mtable


class FairRank(FairEstimator):
    """
    FA*IR: A Fair Top-k Ranking Algorithm

    Reference:
        https://dl.acm.org/doi/pdf/10.1145/3132847.3132938
    Code adopted from:
        https://github.com/fair-search/fairsearch-fair-python

    Attributes
    ----------
    s_classes_ : ndarray of shape (n_sensitive_group,)
        A list of sensitive classes known to LabelBias during training.

    Examples
    --------
    >>> from fairpy.model import FairRank
    >>> model = FairRank(K=5, P=0.5, alpha=0.10)
    >>> scores = [0.98, 0.97, 0.85, 0.84, 0.83, 0.55]
    >>> s = ["male", "male", "male", "female", "female", "female"]
    >>> fair_rank = model.transform(scores=scores, s=s)
    """

    _parameter_constraints = {
        "K": [Interval(Integral, 10, 400, closed="both")],
        "P": [Interval(Real, 0.02, 0.98, closed="both")],
        "alpha": [Interval(Real, 0.01, 0.15, closed="both")],
    }

    def __init__(self, K: int, P: float, alpha: float):
        """
        Parameters
        ----------
        K : int
            Number of Top-K elements returned, should be in range [10, 400].

        P : float
            Proportion of sensitive attributes in the Top-K elements, should be in range [0.02, 0.98].

        alpha : float
            Significance level, should be in range [0.01, 0.15].
        """

        self.K = K
        self.P = P
        self.alpha = alpha

    def _create_mtable(self, alpha: float, adjust_alpha: bool) -> List:
        fc = MTableGenerator(self.K, self.P, alpha, adjust_alpha)
        return fc.mtable_as_list()

    def transform(self, scores: npt.ArrayLike, s: npt.ArrayLike) -> List[int]:
        """
        Transform the ranking to be fair based on scores.

        Parameters
        ----------
        scores : array-like of shape (n_samples,)
            Scores for ranking samples, where `n_samples` is the number of samples.

        s : array-like of shape (n_samples,)
            Sensitive attributes relative to scores.

        Returns
        -------
        rank : array-like of shape (n_samples,)
            Fair rank with indexes corresponding to scores.
        """

        self._validate_params()
        scores, s = self._validate_data(scores, s)
        s = self._validate_grp_s(s)

        mtable = self._create_mtable(self.alpha, True)

        protected_idx_list = np.where(s == 0)[0]
        non_protected_idx_list = np.where(s == 1)[0]
        N_protected = len(protected_idx_list)
        N_non_protected = len(non_protected_idx_list)

        idx_protected = 0
        idx_non_protected = 0
        res = []
        for i in range(self.K):
            if idx_protected >= N_protected and idx_non_protected >= N_non_protected:
                # no more candidate available
                return res

            elif idx_protected >= N_protected:
                # no more protected candidate available, take non-protected candidate
                res.append(non_protected_idx_list[idx_non_protected])
                idx_non_protected += 1

            elif idx_non_protected >= N_non_protected:
                # no more non-protected candidate available, take protected candidate
                res.append(protected_idx_list[idx_protected])
                idx_protected += 1

            elif idx_protected < mtable[i]:
                # add a protected candidate
                res.append(protected_idx_list[idx_protected])
                idx_protected += 1
            else:
                # find the best candidate available
                if scores[protected_idx_list[idx_protected]] >= scores[non_protected_idx_list[idx_non_protected]]:
                    res.append(protected_idx_list[idx_protected])
                    idx_protected += 1
                else:
                    res.append(non_protected_idx_list[idx_non_protected])
                    idx_non_protected += 1

        return res

    def _more_tags(self):
        return {
            'requires_s': True,
            "s_types": ["binary"],
        }



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import numpy as np
import numpy.typing as npt

from sklearn.metrics._classification import column_or_1d, confusion_matrix
from sklearn.utils.multiclass import type_of_target
from sklearn.utils.validation import check_consistent_length


def binary_dp(y_pred: npt.ArrayLike, s: npt.ArrayLike, abs: bool = True) -> float:
    """
    Demographic parity for binary sensitive attribute and binary classification

    Parameters
    ----------
    y_pred : 1d array-like
        Predicted labels, as returned by a classifier.

    s : 1d array-like
        Sensitive attributes.

    abs : bool, default=True
        If ``True``, return the absolute value of dp.
        Otherwise, treat sensitive attribute with value 0 as the privileged group

    Returns
    -------
    dp : float
        Return the gap in demographic parity between two sensitive groups.
        The ideal performance for fairness is 0.

    Examples
    --------

    """

    check_consistent_length(y_pred, s)
    y_type = type_of_target(y_pred)
    s_type = type_of_target(s)

    if y_type != "binary":
        raise ValueError("binary_dp does not support {0} predictions".format(y_type))
    if s_type != "binary":
        raise ValueError("binary_dp does not support {0} sensitive attribute".format(y_type))
    if len(np.unique(s)) == 1:
        raise ValueError("Demographic Parity does not support only one sensitive group")

    y_pred = column_or_1d(y_pred)
    s = column_or_1d(s)

    s_val = np.unique(s)
    grp_0_idx = np.where(s == s_val[0])
    grp_1_idx = np.where(s == s_val[1])

    dp = np.mean(y_pred.take(grp_0_idx)) - np.mean(y_pred.take(grp_1_idx))
    if abs:
        dp = np.abs(dp)

    return dp


def binary_eop(
        y_true: npt.ArrayLike,
        y_pred: npt.ArrayLike,
        s: npt.ArrayLike,
        abs: bool = True,
        labels: npt.ArrayLike = None,
) -> float:
    """
    Equal Opportunity for binary sensitive attribute and binary classification

    Parameters
    ----------
    y_true : 1d array-like
        Ground truth (correct) labels.

    y_pred : 1d array-like
        Predicted labels, as returned by a classifier.

    s : 1d array-like
        Sensitive attributes.

    abs : bool, default=True
        If ``True``, return the absolute value of dp.
        Otherwise, treat sensitive attribute with value 0 as the privileged group

    labels : array-like of shape (n_classes), default=None
        List of labels to index the classes.

    Returns
    -------
    eop : float
        Return the gap in true positive rates between two sensitive groups.
        The ideal performance for fairness is 0.

    Examples
    --------

    """

    check_consistent_length(y_true, y_pred, s)
    y_true_type = type_of_target(y_true)
    y_pred_type = type_of_target(y_pred)
    s_type = type_of_target(s)

    if y_true_type != "binary":
        raise ValueError("binary_eop does not support {0} targets".format(y_true_type))
    if y_pred_type not in ("binary", "multiclass"):
        raise ValueError("binary_eop does not support {0} predictions".format(y_pred_type))
    elif y_pred_type == "multiclass":
        raise Warning("Expected binary classification, but predictions contain more than two classes")
    if s_type != "binary":
        raise ValueError("binary_eop does not support {0} sensitive attribute".format(s_type))
    if len(np.unique(s)) == 1:
        raise ValueError("Equal Opportunity does not support only one sensitive group")

    y_true = column_or_1d(y_true)
    y_pred = column_or_1d(y_pred)
    s = column_or_1d(s)

    s_val = np.unique(s)
    grp_0_idx = np.where(s == s_val[0])
    grp_1_idx = np.where(s == s_val[1])

    cm_grp_0 = confusion_matrix(y_true[grp_0_idx].squeeze(), y_pred[grp_0_idx].squeeze(), labels=labels)
    cm_grp_1 = confusion_matrix(y_true[grp_1_idx].squeeze(), y_pred[grp_1_idx].squeeze(), labels=labels)

    tpr_grp_0 = cm_grp_0[1][1] / (cm_grp_0[1][1] + cm_grp_0[1][0])
    tpr_grp_1 = cm_grp_1[1][1] / (cm_grp_1[1][1] + cm_grp_1[1][0])
    eop = tpr_grp_0 - tpr_grp_1

    if abs:
        eop = np.abs(eop)

    return eop


def binary_eod(y_true: npt.ArrayLike, y_pred: npt.ArrayLike, s: npt.ArrayLike, labels: npt.ArrayLike = None):
    """
    Equal Odds for binary sensitive attribute and binary classification

    Parameters
    ----------
    y_true : 1d array-like
        Ground truth (correct) labels.

    y_pred : 1d array-like
        Predicted labels, as returned by a classifier.

    s : 1d array-like
        Sensitive attributes.

    abs : bool, default=True
        If ``True``, return the absolute value of dp.
        Otherwise, treat sensitive attribute with value 0 as the privileged group

    labels : array-like of shape (n_classes), default=None
        List of labels to index the classes.

    Returns
    -------
    eod : float
        Return sum of the gap in true positive rate and false positive rate between two sensitive groups.
        The ideal performance for fairness is 0.

    Examples
    --------

    """

    check_consistent_length(y_true, y_pred, s)
    y_true_type = type_of_target(y_true)
    y_pred_type = type_of_target(y_pred)
    s_type = type_of_target(s)

    if y_true_type != "binary":
        raise ValueError("binary_eop does not support {0} targets".format(y_true_type))
    if y_pred_type not in ("binary", "multiclass"):
        raise ValueError("binary_eop does not support {0} predictions".format(y_pred_type))
    elif y_pred_type == "multiclass":
        raise Warning("Expected binary classification, but predictions contain more than two classes")
    if s_type != "binary":
        raise ValueError("binary_eop does not support {0} sensitive attribute".format(s_type))
    if len(np.unique(s)) == 1:
        raise ValueError("Equal Opportunity does not support only one sensitive group")

    y_true = column_or_1d(y_true)
    y_pred = column_or_1d(y_pred)
    s = column_or_1d(s)

    s_val = np.unique(s)
    grp_0_idx = np.where(s == s_val[0])
    grp_1_idx = np.where(s == s_val[1])

    cm_grp_0 = confusion_matrix(y_true[grp_0_idx].squeeze(), y_pred[grp_0_idx].squeeze(), labels=labels)
    cm_grp_1 = confusion_matrix(y_true[grp_1_idx].squeeze(), y_pred[grp_1_idx].squeeze(), labels=labels)

    tpr_grp_0 = cm_grp_0[1][1] / (cm_grp_0[1][1] + cm_grp_0[1][0])
    tpr_grp_1 = cm_grp_1[1][1] / (cm_grp_1[1][1] + cm_grp_1[1][0])
    fpr_grp_0 = cm_grp_0[0][1] / (cm_grp_0[0][1] + cm_grp_0[0][0])
    fpr_grp_1 = cm_grp_1[0][1] / (cm_grp_1[0][1] + cm_grp_1[0][0])

    eod = abs(tpr_grp_0 - tpr_grp_1) + abs(fpr_grp_0 - fpr_grp_1)

    return eod



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import os
import sys
import pandas as pd
from typing import Sequence, Union

from ._base import TabDataset


class German(TabDataset):
    """ https://archive.ics.uci.edu/ml/datasets/Statlog+%28German+Credit+Data%29 """

    file2url = {
        "german.data": "https://archive.ics.uci.edu/ml/machine-learning-databases/statlog/german/german.data",
    }

    column_names = (
        "status",
        "month",
        "credit_history",
        "purpose",
        "credit_amount",
        "savings",
        "employment",
        "investment_as_income_percentage",
        "personal_status",
        "other_debtors",
        "residence_since",
        "property",
        "age",
        "installment_plans",
        "housing",
        "number_of_credits",
        "skill_level",
        "people_liable_for",
        "telephone",
        "foreign_worker",
        "credit",
    )

    categorical_feat = [
        "status",
        "credit_history",
        "purpose",
        "savings",
        "employment",
        "other_debtors",
        "property",
        "installment_plans",
        "housing",
        "skill_level",
        "telephone",
        "foreign_worker",
        "sex",
        "marital-status",
    ]

    numerical_feat = [
        "month",
        "credit_amount",
        "investment_as_income_percentage",
        "residence_since",
        "age",
        "number_of_credits",
        "people_liable_for",
    ]

    drop_feat = ["personal_status"]
    na_values = ()

    label_name = "credit"
    label_mapping = ("2", "1")

    dataset_name = "german"

    def __init__(
            self,
            sen_feat: Union[str, Sequence[str]] = "sex",
            dir_path: str = os.path.join(sys.path[1], "fairpy/data/german"),
            download: bool = False,
    ):
        self._check_and_download_data(dir_path, self.file2url, download)

        df = pd.read_csv(
            os.path.join(dir_path, "german.data"),
            sep=" ",
            names=self.column_names,
        )

        super(German, self).__init__(
            df=df,
            label_name=self.label_name,
            sen_feat=sen_feat,
            cat_feat=self.categorical_feat,
            num_feat=self.numerical_feat,
            dataset_name=self.dataset_name,
            custom_preprocessing=self.default_preprocessing,
            na_values=self.na_values,
            drop_feat=self.drop_feat,
            label_mapping=self.label_mapping,
        )

    @staticmethod
    def default_preprocessing(df: pd.DataFrame) -> pd.DataFrame:
        """
        Adds a derived sex attribute based on personal_status.
        https://github.com/Trusted-AI/AIF360/blob/master/aif360/datasets/german_dataset.py
        """

        sex_map = {'A91': 'male', 'A93': 'male', 'A94': 'male', 'A92': 'female', 'A95': 'female'}
        marital_status_map = {'A91': 'married', 'A92': 'married', 'A93': 'single', 'A94': 'married', 'A95': 'single'}
        df['sex'] = df['personal_status'].replace(sex_map)
        df["marital-status"] = df["personal_status"].replace(marital_status_map)
        df["credit"] = df["credit"].astype(str)

        return df



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

from . import dataset
from . import model
from . import metric
from . import utils

__all__ = [
    "dataset",
    "model",
    "metric",
]



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import os
import gdown
import tarfile
import pandas as pd
import numpy as np
import numpy.typing as npt
from zipfile import ZipFile
from collections import OrderedDict
from typing import Sequence, Any, Callable, Optional, NamedTuple, Tuple, Union, Mapping
from abc import ABC, abstractmethod

from sklearn.preprocessing import OneHotEncoder, LabelEncoder
from sklearn.model_selection import train_test_split


class FeatIndex(NamedTuple):
    """ Feature names and indexes after one-hot encoding for tabular dataset """
    cat_feat: Tuple[str, ...]
    num_feat: Tuple[str, ...]
    sen_feat: Tuple[str, ...]
    cat_idx: Tuple[int, ...]
    num_idx: Tuple[int, ...]
    sen_idx: Tuple[int, ...]
    feat2idx: Mapping[str, Tuple[int, ...]]


class SplitData(NamedTuple):
    """ Training data, label, and sensitive attributes after train, validation, and test split """
    X_train: npt.NDArray[np.float64]
    y_train: np.ndarray
    s_train: npt.NDArray[np.int16]
    X_val: npt.NDArray[np.float64]
    y_val: np.ndarray
    s_val: npt.NDArray[np.int16]
    X_test: npt.NDArray[np.float64]
    y_test: np.ndarray
    s_test: npt.NDArray[np.int16]


class Dataset(ABC):
    """ Abstract base class for all datasets """

    def _check_and_download_data(
            self,
            dir_path: str,
            file2url: Mapping[str, str],
            download: bool,
    ) -> None:
        """
        Check local data and download if no data available

        Parameters
        ----------
        dir_path: str
            Directory of the data folder

        file2url: str -> str
            File name and corresponding url for downloading

        download: bool
            Set to True to enable automatic downloading

        use_gdown: bool
            Set to True to use gdown for downloading

        """

        def check_file_exist(dir_path: str, file2url: Mapping[str, str]) -> Mapping[str, bool]:
            """ Existence flag for each file """
            flag = {f: False for f in file2url.keys()}
            for file_name in file2url.keys():
                if os.path.isfile(os.path.join(dir_path, file_name)):
                    flag[file_name] = True
            return flag

        def raise_not_exist(exist_flag: Mapping[str, bool]) -> None:
            for file_name, exist in exist_flag.items():
                if not exist:
                    raise FileExistsError(
                        "%s does not exist in %s, consider set `download` to Ture" % (file_name, dir_path)
                    )

            return

        def download_data(
                dir_path: str,
                file2url: Mapping[str, str],
                exist_flag: Mapping[str, bool],
        ) -> None:
            if not os.path.exists(dir_path):
                os.makedirs(dir_path)

            for file_name, url in file2url.items():
                if not exist_flag[file_name]:
                    print("FairPy => Download %s from %s" % (file_name, url))
                    gdown.download(url, os.path.join(dir_path, file_name), quiet=False)

                # TODO: include all compressed file
                if file_name.endswith(".zip"):
                    zf = ZipFile(os.path.join(dir_path, file_name), 'r')
                    zf.extractall(dir_path)
                    zf.close()
                elif file_name.endswith(".tar.gz"):
                    f = tarfile.open(os.path.join(dir_path, file_name))
                    f.extractall(dir_path)
                    f.close()

            return

        exist_flag = check_file_exist(dir_path, file2url)
        if not download:
            raise_not_exist(exist_flag)
        else:
            download_data(dir_path, file2url, exist_flag)

        return


class TabDataset(Dataset):
    """
    Base class for fair classification on tabular datasets

    Parameters
    ----------
    df: pd.DataFrame
        Tabular dataset in pandas dataframe

    label_name: str
        The name of target variable in dataframe

    sen_feat: str or Sequence[str]
        The name of sensitive feature(s), can be one or multiple features

    cat_feat: Sequence[str]
        The list of names of categorical features

    num_feat: Sequence[str]
        The list of names of numerical features

    dataset_name: str
        The name of the dataset

    custom_preprocessing: Callable: pd.DataFrame -> pd.DataFrame
        Customized preprocessing for dataframe, e.g., data selection or deletion

    """

    def __init__(
            self,
            df: pd.DataFrame,
            label_name: str,
            sen_feat: Union[str, Sequence[str]],
            cat_feat: Sequence[str],
            num_feat: Sequence[str],
            dataset_name: Optional[str] = None,
            custom_preprocessing: Optional[Callable[[pd.DataFrame], pd.DataFrame]] = None,
            na_values: Optional[Sequence[Any]] = None,
            drop_feat: Optional[Sequence[str]] = None,
            label_mapping: Optional[Sequence[Any]] = None,
    ):
        if isinstance(sen_feat, str):
            sen_feat = tuple([sen_feat])

        self.df = df
        self.label_name = label_name
        if isinstance(sen_feat, str):
            self._sen_feat = tuple([sen_feat])
        else:
            self._sen_feat = tuple(sen_feat)
        self._cat_feat = tuple(cat_feat)
        self._num_feat = tuple(num_feat)
        self.dataset_name = dataset_name
        self.na_values = na_values
        self.drop_feat = drop_feat
        self.label_mapping = label_mapping

        self._feat2idx = None
        self.label_encoder = LabelEncoder()
        self.sen_encoder = LabelEncoder()
        if self.label_mapping is not None:
            self.label_encoder.classes_ = self.label_mapping

        if custom_preprocessing is not None:
            self.df = custom_preprocessing(self.df)
        self._std_preprocessing()

        self.input_df = self.df[[feat for feat in self._num_feat] + [feat for feat in self._cat_feat]]
        self.target_df = self.df[self.label_name]
        self.sen_df = self.df[list(self._sen_feat)]

        # TODO: more encoding?
        self.encoder = OneHotEncoder(sparse=False, handle_unknown="ignore")
        self.encoded_df = self.one_hot_encoding(self.df)

    def _std_preprocessing(self) -> None:
        """ Drop features and delete row with nan values """

        if self.drop_feat is not None:
            self.df.drop(columns=self.drop_feat, inplace=True)
        if self.na_values is not None:
            self.df.replace(self.na_values, np.nan, inplace=True)
            self.df.dropna(axis=0, inplace=True)
            self.df.reset_index(inplace=True)

        return

    def one_hot_encoding(self, df: pd.DataFrame) -> pd.DataFrame:
        """ One-hot encoding for categorical features """

        df_cat = df[list(self._cat_feat)]
        df_num = df[list(self._num_feat)]

        if hasattr(self.encoder, "categories_"):
            df_cat_values = self.encoder.transform(df_cat.values)
        else:
            df_cat_values = self.encoder.fit_transform(df_cat.values)

            self._cat_col_name = ["%s.%s" % (name, val) for name, all_val in
                                  zip(self._cat_feat, self.encoder.categories_) for val in all_val]
            self._feat2idx = OrderedDict({feat: [i] for i, feat in enumerate(self._num_feat)})
            curr_idx = len(self._num_feat)
            for i, feat in enumerate(self._cat_feat):
                feat_idx = [j + curr_idx for j in range(len(self.encoder.categories_[i]))]
                self._feat2idx.update({feat: feat_idx})
                curr_idx += len(feat_idx)

            self._cat_idx, self._num_idx, self._sen_idx = [], [], []
            for feat in self._feat2idx.keys():
                if feat in self._cat_feat:
                    self._cat_idx.extend(self._feat2idx[feat])
                if feat in self._num_feat:
                    self._num_idx.extend(self._feat2idx[feat])
                if feat in self._sen_feat:
                    self._sen_idx.extend(self._feat2idx[feat])

            self._cat_idx = tuple(self._cat_idx)
            self._num_idx = tuple(self._num_idx)
            self._sen_idx = tuple(self._sen_idx)
            self._feat2idx = {key: tuple(list_) for key, list_ in self._feat2idx.items()}
            self._feat2idx = OrderedDict(sorted(self._feat2idx.items(), key=lambda x: x[1][0]))  # order by index

        df_cat = pd.DataFrame(df_cat_values, columns=self._cat_col_name)
        encoded_df = pd.concat([df_num, df_cat], axis=1)

        return encoded_df

    @property
    def feat_idx(self) -> FeatIndex:
        cat_idx = tuple([idx for feat in self._cat_feat for idx in self._feat2idx[feat]])
        num_idx = tuple([idx for feat in self._num_feat for idx in self._feat2idx[feat]])
        sen_idx = tuple([idx for feat in self._sen_feat for idx in self._feat2idx[feat]])

        return FeatIndex(self._cat_feat, self._num_feat, self._sen_feat, cat_idx, num_idx, sen_idx, self._feat2idx)

    def get_X_y_s(
            self,
            onehotify: bool = True,
    ) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.int16], npt.NDArray[np.int16]]:
        """ Return training data, labels, and sensitive attributes """

        y = self.target_df.to_numpy()
        if hasattr(self.label_encoder, "classes_"):
            y = self.label_encoder.transform(y)
        else:
            y = self.label_encoder.fit_transform(y)

        # TODO: add support for multiple sensitive attribute
        s = self.sen_df.to_numpy().squeeze()
        s = self.sen_encoder.fit_transform(s)

        if onehotify:
            X = self.encoded_df.to_numpy()
            X, y, s = self.validate(X, y, s)
        else:
            X = self.input_df.to_numpy()

        return X, y, s

    def split(self, val_p: float = 0.15, test_p: float = 0.15, random_state: int = 42):
        assert val_p < 1. and test_p < 1., "Invalid partition of data"

        X, y, s = self.get_X_y_s()
        X_train, X_test, y_train, y_test, s_train, s_test = train_test_split(
            X, y, s, test_size=test_p, random_state=random_state)
        if val_p != 0.:
            X_train, X_val, y_train, y_val, s_train, s_val = train_test_split(
                X_train, y_train, s_train, test_size=val_p / (1 - test_p), random_state=random_state)
        else:
            X_val = y_val = s_val = None

        return SplitData(X_train, y_train, s_train, X_val, y_val, s_val, X_test, y_test, s_test)

    def validate(self, X: np.ndarray, y: np.ndarray, s: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        X = X.astype(np.float64)
        y = y.astype(np.int16)
        s = s.astype(np.int16)

        assert len(np.unique(y)) > 1, "Only one class exists in the dataset"
        assert np.any(y) >= 0, "Class label should be an integer larger or equal to zero"
        if len(np.unique(s)) == 1:
            import warnings
            warnings.warn("Only one sensitive group in the dataset")

        return X, y, s

    def get_params(self):
        return

    @property
    def name(self):
        return self.dataset_name



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import os
import sys
import pandas as pd
from typing import Sequence, Union

from ._base import TabDataset


class Adult(TabDataset):
    """ https://archive.ics.uci.edu/ml/datasets/adult """

    file2url = {
        "adult.data": "https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.data",
        "adult.test": "https://archive.ics.uci.edu/ml/machine-learning-databases/adult/adult.test",
    }

    column_names = (
        'age',
        'workclass',
        'fnlwgt',
        'education',
        'education-num',
        'marital-status',
        'occupation',
        'relationship',
        'race',
        'sex',
        'capital-gain',
        'capital-loss',
        'hours-per-week',
        'native-country',
        'income-per-year',
    )

    categorical_feat = [
        'workclass',
        'education',
        'marital-status',
        'occupation',
        'relationship',
        'race',
        'sex',
        'native-country',
    ]

    numerical_feat = [
        'age',
        'education-num',
        'capital-gain',
        'capital-loss',
        'hours-per-week',
    ]

    drop_feat = ["fnlwgt"]
    na_values = ("?")

    label_name = "income-per-year"
    label_mapping = ("<=50K", ">50K")

    dataset_name = "adult"

    def __init__(
            self,
            sen_feat: Union[str, Sequence[str]] = "sex",
            dir_path: str = os.path.join(sys.path[1], "fairpy/data/adult"),
            download: bool = False,
    ):
        self._check_and_download_data(dir_path, self.file2url, download)

        train_df = pd.read_csv(
            os.path.join(dir_path, "adult.data"),
            header=None,
            names=self.column_names,
            skipinitialspace=True,
        )

        test_df = pd.read_csv(
            os.path.join(dir_path, "adult.test"),
            header=None,
            names=self.column_names,
            skipinitialspace=True,
        )
        for i, row in test_df.iterrows():
            if isinstance(row["income-per-year"], str):
                test_df.at[i, "income-per-year"] = row["income-per-year"][:-1]

        df = pd.concat([train_df, test_df], axis=0)
        df.reset_index(inplace=True)

        super(Adult, self).__init__(
            df=df,
            label_name=self.label_name,
            sen_feat=sen_feat,
            cat_feat=self.categorical_feat,
            num_feat=self.numerical_feat,
            dataset_name=self.dataset_name,
            custom_preprocessing=None,
            na_values=self.na_values,
            drop_feat=self.drop_feat,
            label_mapping=self.label_mapping,
        )



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import math
import numpy as np
import numpy.typing as npt
from typing import List, Tuple

from sklearn.metrics._classification import column_or_1d
from sklearn.utils.multiclass import type_of_target
from sklearn.utils.validation import check_consistent_length


def _fast_auc(y_true, y_pred):
    N = len(y_true)

    y_true = y_true[np.argsort(y_pred)]
    N_false = 0
    auc = 0

    for i in range(N):
        y_i = y_true[i]
        N_false += (1 - y_i)
        auc += y_i * N_false

    auc /= (N_false * (N - N_false))

    return auc


def _cross_auc(pred_1, pred_2):
    scores = np.hstack(np.asarray([pred_1, pred_2]))
    y_true = np.zeros(len(pred_1) + len(pred_2))
    y_true[0:len(pred_1)] = 1

    return _fast_auc(y_true, scores)


def xAUC(y_true: npt.ArrayLike, y_pred: npt.ArrayLike, s: npt.ArrayLike, abs: bool = False) -> float:
    """
    The Fairness of Risk Scores Beyond Classification: Bipartite Ranking and the xAUC Metric

    Parameters
    ----------
    y_true : 1d array-like
        Ground truth (correct) labels.

    y_pred : 1d array-like
        Predicted labels, as returned by a classifier.

    s : 1d array-like
        Sensitive attributes.

    abs : bool, default=True
        If ``True``, return the absolute value of dp.
        Otherwise, treat sensitive attribute with value 0 as the privileged group

    Returns
    -------
    xAUC : float
        Disparity in the probabilities of ranking a positive example from one group above a negative one
        from another group.
        The best performance is 0.

    Examples
    --------
    from fairpy.metric import xAUC

    y_true = [0, 0, 1, 0, 1, 1, 0, 1]
    y_pred = [0.1, 0.2, 0.3, 0.8, 0.9, 1.5, 0, 1]
    s = ["male", "female", "male", "female", "male", "male", "male", "female"]

    xauc = xAUC(y_true, y_pred, s)
    """

    check_consistent_length(y_true, y_pred, s)
    y_true_type = type_of_target(y_true)
    y_pred_type = type_of_target(y_pred)
    s_type = type_of_target(s)

    if y_true_type != "binary":
        raise ValueError("xAUC does not support {0} targets".format(y_true_type))
    if y_pred_type not in ("binary", "continuous"):
        raise ValueError("xAUC does not support {0} predictions".format(y_pred_type))

    if s_type != "binary":
        raise ValueError("xAUC does not support {0} sensitive attribute".format(s_type))
    if len(np.unique(s)) == 1:
        raise ValueError("xAUC does not support only one sensitive group")

    y_true = column_or_1d(y_true)
    y_pred = column_or_1d(y_pred)
    s = column_or_1d(s)

    s_val = np.unique(s)
    grp_0_idx = np.where(s == s_val[0])
    grp_1_idx = np.where(s == s_val[1])
    y_0_idx = np.where(y_true == 0)
    y_1_idx = np.where(y_true == 1)

    grp_0_y_0_cross_grp_1_y_1 = _cross_auc(
        y_pred[np.intersect1d(grp_0_idx, y_0_idx)],
        y_pred[np.intersect1d(grp_1_idx, y_1_idx)]
    )

    grp_1_y_0_cross_grp_0_y_1 = _cross_auc(
        y_pred[np.intersect1d(grp_1_idx, y_0_idx)],
        y_pred[np.intersect1d(grp_0_idx, y_1_idx)]
    )

    dp = grp_0_y_0_cross_grp_1_y_1 - grp_1_y_0_cross_grp_0_y_1
    if abs:
        dp = np.abs(dp)

    return dp


def dcg(ranking: List[int], rel_vec: npt.ArrayLike, k: int = 0) -> Tuple[float, float]:
    """ Discounted cumulative gain (dcg) and normalized discounted cumulative gain (ndcg) """

    dcgmax = 0.
    dcg = 0.
    sorted_rel = -np.sort(-rel_vec)
    N = len(rel_vec)

    if k == 0:
        k = N

    for i, rel in enumerate(sorted_rel[:min((k, N))]):
        dcgmax += float(2.0 ** rel - 1) / math.log2(2 + i)
    for i, doc in enumerate(ranking[:min(k, N)]):
        dcg += float(2 ** rel_vec[doc] - 1) / math.log2(2 + i)

    if dcgmax == 0:
        return 1.0, 1.0
    else:
        return dcg / dcgmax, dcg



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

from ._classification import binary_dp
from ._classification import binary_eop
from ._ranking import xAUC, dcg

__all__ = [
    "binary_dp",
    "binary_eop",
    "xAUC",
    "dcg",
]



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

from __future__ import annotations

import copy
import numpy as np
import numpy.typing as npt

import cvxpy as cvx
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.validation import _check_y, check_consistent_length

from ._base import FairEstimator


class EqOddsCalib(FairEstimator):
    """
    Equality of Opportunity in Supervised Learning

    Reference:
        https://arxiv.org/pdf/1610.02413.pdf
    Code adopted from:
        https://github.com/gpleiss/equalized_odds_and_calibration

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        A list of class labels known to the classifier.

    s_classes_ : ndarray of shape (n_sensitive_group,)
        A list of sensitive classes known to LabelBias during training.

    n_features_in_ : int
        Number of features seen during `fit`.

    feature_names_in_ : ndarray of shape (n_features_in_,)
        Names of features seen during `fit`. Defined only when X has feature names that are all strings.

    Examples
    --------
    >>> from fairpy.dataset import Adult
    >>> from fairpy.model import EqOddsCalib
    >>> dataset = Adult()
    >>> split_data = dataset.split()
    >>> EqOddsCalib.fit(pred=pred, y=split_data.y_train, s=split_data.s_train)
    >>> fair_pred = EqOddsCalib.transform(pred=pred, s=split_data.s_test)
    """

    @staticmethod
    def base_rate(label):
        """
        Percentage of samples belonging to the positive class
        """
        return np.mean(label)

    @staticmethod
    def tpr(pred, label):
        """
        True positive rate
        """
        return np.mean(np.logical_and(pred.round() == 1, label == 1))

    @staticmethod
    def fpr(pred, label):
        """
        False positive rate
        """
        return np.mean(np.logical_and(pred.round() == 1, label == 0))

    @staticmethod
    def tnr(pred, label):
        """
        True negative rate
        """
        return np.mean(np.logical_and(pred.round() == 0, label == 0))

    @staticmethod
    def fnr(pred, label):
        """
        False negative rate
        """
        return np.mean(np.logical_and(pred.round() == 0, label == 1))

    @staticmethod
    def fn_cost(pred, label):
        """
        Generalized false negative cost
        """
        return 1 - pred[label == 1].mean()

    @staticmethod
    def fp_cost(pred, label):
        """
        Generalized false positive cost
        """
        return pred[label == 0].mean()

    def fit(self, pred: npt.ArrayLike, y: npt.ArrayLike, s: npt.ArrayLike) -> EqOddsCalib:
        """
        Fit the model according to the given training data.

        Parameters
        ----------
        pred : array-like of shape (n_samples,)
            Predictions to be calibrated, where `n_samples` is the number of samples.

        y : array-like of shape (n_samples,)
            Target vector relative to predictions.

        s : array-like of shape (n_samples,)
            Sensitive attributes relative to X.

        Returns
        -------
        self
            Fitted estimator.
        """

        # TODO: check the implementation in AIF360

        y, s = self._validate_data(y=y, s=s)
        pred = _check_y(pred)
        check_consistent_length(pred, y)
        y = self._validate_cls_y(y)
        s = self._validate_grp_s(s)

        if np.max(pred) > 1:
            import warnings
            from sklearn.exceptions import DataConversionWarning
            warnings.warn("Predictions have values larger than 1, expected in the range of [0, 1]",
                          DataConversionWarning)
        if np.min(pred) < 0:
            import warnings
            from sklearn.exceptions import DataConversionWarning
            warnings.warn("Predictions have values smaller than 0, expected in the range of [0, 1]",
                          DataConversionWarning)

        grp_0_idx, grp_1_idx = np.where(s == 0)[0], np.where(s == 1)[0]
        pred_grp_0, pred_grp_1 = np.take(pred, grp_0_idx), np.take(pred, grp_1_idx)
        y_grp_0, y_grp_1 = np.take(y, grp_0_idx), np.take(y, grp_1_idx)

        sbr = self.base_rate(y_grp_0)
        obr = self.base_rate(y_grp_1)

        sp2p = cvx.Variable(1)
        sp2n = cvx.Variable(1)
        sn2p = cvx.Variable(1)
        sn2n = cvx.Variable(1)

        op2p = cvx.Variable(1)
        op2n = cvx.Variable(1)
        on2p = cvx.Variable(1)
        on2n = cvx.Variable(1)

        sfpr = self.fpr(pred_grp_0, y_grp_0) * sp2p + self.tnr(pred_grp_0, y_grp_0) * sn2p
        sfnr = self.fnr(pred_grp_0, y_grp_0) * sn2n + self.tpr(pred_grp_0, y_grp_0) * sp2n
        ofpr = self.fpr(pred_grp_1, y_grp_1) * op2p + self.tnr(pred_grp_1, y_grp_1) * on2p
        ofnr = self.fnr(pred_grp_1, y_grp_1) * on2n + self.tpr(pred_grp_1, y_grp_1) * op2n
        error = sfpr + sfnr + ofpr + ofnr

        sflip = 1 - pred_grp_0
        sconst = pred_grp_0
        oflip = 1 - pred_grp_1
        oconst = pred_grp_1

        sm_tn = np.logical_and(pred_grp_0 == 0, y_grp_0 == 0)
        sm_fn = np.logical_and(pred_grp_0 == 0, y_grp_0 == 1)
        sm_tp = np.logical_and(pred_grp_0 == 1, y_grp_0 == 1)
        sm_fp = np.logical_and(pred_grp_0 == 1, y_grp_0 == 0)

        om_tn = np.logical_and(pred_grp_1 == 0, y_grp_1 == 0)
        om_fn = np.logical_and(pred_grp_1 == 0, y_grp_1 == 1)
        om_tp = np.logical_and(pred_grp_1 == 1, y_grp_1 == 1)
        om_fp = np.logical_and(pred_grp_1 == 1, y_grp_1 == 0)

        spn_given_p = (sn2p * (sflip * sm_fn).mean() + sn2n * (sconst * sm_fn).mean()) / sbr + \
                      (sp2p * (sconst * sm_tp).mean() + sp2n * (sflip * sm_tp).mean()) / sbr

        spp_given_n = (sp2n * (sflip * sm_fp).mean() + sp2p * (sconst * sm_fp).mean()) / (1 - sbr) + \
                      (sn2p * (sflip * sm_tn).mean() + sn2n * (sconst * sm_tn).mean()) / (1 - sbr)

        opn_given_p = (on2p * (oflip * om_fn).mean() + on2n * (oconst * om_fn).mean()) / obr + \
                      (op2p * (oconst * om_tp).mean() + op2n * (oflip * om_tp).mean()) / obr

        opp_given_n = (op2n * (oflip * om_fp).mean() + op2p * (oconst * om_fp).mean()) / (1 - obr) + \
                      (on2p * (oflip * om_tn).mean() + on2n * (oconst * om_tn).mean()) / (1 - obr)

        constraints = [
            sp2p == 1 - sp2n,
            sn2p == 1 - sn2n,
            op2p == 1 - op2n,
            on2p == 1 - on2n,
            sp2p <= 1,
            sp2p >= 0,
            sn2p <= 1,
            sn2p >= 0,
            op2p <= 1,
            op2p >= 0,
            on2p <= 1,
            on2p >= 0,
            spp_given_n == opp_given_n,
            spn_given_p == opn_given_p,
        ]

        prob = cvx.Problem(cvx.Minimize(error), constraints)
        prob.solve()

        self._sp2p = sp2p.value
        self._sn2p = sn2p.value
        self._op2p = op2p.value
        self._on2p = on2p.value

        return self

    def transform(self, pred: npt.ArrayLike, s: npt.ArrayLike) -> npt.NDArray:
        """
        Transform predictions to be fair.

        Parameters
        ----------
        pred : array-like of shape (n_samples,)
            Predictions to be calibrated, where `n_samples` is the number of samples.

        s : array-like of shape (n_samples,)
            Sensitive attributes relative to X.

        Returns
        -------
        pred : ndarray of shape (n_samples,)
            Debiased predictions.
        """

        check_is_fitted(self)
        pred, s = self._validate_data(y=pred, s=s, reset=False)
        s = self._validate_grp_s(s)

        fair_pred = copy.deepcopy(pred)
        grp_0_idx, grp_1_idx = np.where(s == 0)[0], np.where(s == 1)[0]

        pp_indices = np.intersect1d(np.where(pred == 1)[0], grp_0_idx)
        pn_indices = np.intersect1d(np.where(pred == 0)[0], grp_0_idx)
        np.random.shuffle(pp_indices)
        np.random.shuffle(pn_indices)

        n2p_indices = pn_indices[:int(len(pn_indices) * self._sn2p)]
        fair_pred[n2p_indices] = 1 - fair_pred[n2p_indices]
        p2n_indices = pp_indices[:int(len(pp_indices) * (1 - self._sp2p))]
        fair_pred[p2n_indices] = 1 - fair_pred[p2n_indices]

        othr_pp_indices = np.intersect1d(np.where(pred == 1)[0], grp_1_idx)
        othr_pn_indices = np.intersect1d(np.where(pred == 0)[0], grp_1_idx)
        np.random.shuffle(othr_pp_indices)
        np.random.shuffle(othr_pn_indices)

        n2p_indices = othr_pn_indices[:int(len(othr_pn_indices) * self._on2p)]
        fair_pred[n2p_indices] = 1 - fair_pred[n2p_indices]
        p2n_indices = othr_pp_indices[:int(len(othr_pp_indices) * (1 - self._op2p))]
        fair_pred[p2n_indices] = 1 - fair_pred[p2n_indices]

        return fair_pred

    def _more_tags(self):
        return {
            'binary_only': True,
            'requires_y': True,
            'requires_s': True,
            "y_types": ["binary"],
            "s_types": ["binary"],
        }



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import os
import sys
import pandas as pd
from typing import Sequence, Union

from ._base import TabDataset


class Credit(TabDataset):
    """ https://archive.ics.uci.edu/ml/datasets/default+of+credit+card+clients """

    file2url = {
        "default of credit card clients.xls": "https://archive.ics.uci.edu/ml/machine-learning-databases/00350/default%20of%20credit%20card%20clients.xls",
    }

    categorical_feat = [
        "SEX",
        "EDUCATION",
        "MARRIAGE",
        "PAY_0",
        "PAY_2",
        "PAY_3",
        "PAY_4",
        "PAY_5",
        "PAY_6",
    ]

    numerical_feat = [
        "LIMIT_BAL",
        "AGE",
        "BILL_AMT1",
        "BILL_AMT2",
        "BILL_AMT3",
        "BILL_AMT4",
        "BILL_AMT5",
        "BILL_AMT6",
        "PAY_AMT1",
        "PAY_AMT2",
        "PAY_AMT3",
        "PAY_AMT4",
        "PAY_AMT5",
        "PAY_AMT6",
    ]

    drop_feat = []
    na_values = ()

    label_name = "default payment next month"
    label_mapping = (0, 1)

    dataset_name = "credit"

    def __init__(
            self,
            sen_feat: Union[str, Sequence[str]] = "SEX",
            dir_path: str = os.path.join(sys.path[1], "fairpy/data/credit"),
            download: bool = False,
    ):
        self._check_and_download_data(dir_path, self.file2url, download)
        df = pd.read_excel(os.path.join(dir_path, "default of credit card clients.xls"), header=1, index_col=0)

        super(Credit, self).__init__(
            df=df,
            label_name=self.label_name,
            sen_feat=sen_feat,
            cat_feat=self.categorical_feat,
            num_feat=self.numerical_feat,
            dataset_name=self.dataset_name,
            na_values=self.na_values,
            drop_feat=self.drop_feat,
            label_mapping=self.label_mapping,
        )



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import numpy as np
import numpy.typing as npt
import pandas as pd
from typing import Union, Tuple

from sklearn.base import BaseEstimator
from sklearn.utils.multiclass import type_of_target
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.validation import check_array, check_X_y, _check_y, check_consistent_length


class FairEstimator(BaseEstimator):
    """
    Abstract class for all algorithms in FairPy
    """

    def _validate_data(
            self,
            X: Union[npt.ArrayLike, pd.DataFrame] = "no_validation",
            y: npt.ArrayLike = "no_validation",
            s: npt.ArrayLike = "no_validation",
            reset: bool = True,
            **check_params,
    ) -> Union[npt.NDArray, Tuple[npt.NDArray, ...]]:
        """
        Override method `_validate_data` in `BaseEstimator` to include sensitive attributes `s`.

        Validate input data and set or check the `n_features_in_` attribute. Currently we do not implement
        `validate_separately` in this function.

        Seven validation cases in total:
            1. only X; 2. X and y; 3. X and s; 4. X and y and s; 5. only y; 6. y and s; 7. only s.

        Parameters
        ----------
        X : {array-like, sparse matrix, dataframe} of shape (n_samples, n_features), default='no validation'
            The input samples.
            If `'no_validation'`, no validation is performed on `X`. This is
            useful for meta-estimator which can delegate input validation to
            their underlying estimator(s). In that case `y` must be passed and
            the only accepted `check_params` are `multi_output` and
            `y_numeric`.

        y : array-like of shape (n_samples,), default='no_validation'
            The targets.

            - If `None`, `check_array` is called on `X` or `s`. If the estimator's
              requires_y tag is True, then an error will be raised.
            - If `'no_validation'`, `check_array` is called on `X` or `s` and the
              estimator's requires_y tag is ignored. This is a default
              placeholder and is never meant to be explicitly set. In that case
              `X` must be passed.
            - Otherwise, only `y` with `_check_y` or both `X` and `y` are
              checked with either `check_array` or `check_X_y` depending on
              `validate_separately`.

        s : array-like of shape (n_samples,), default='no_validation'
            The sensitive attributes.

            - If `None`, `check_array` is called on `X` or `y`. If the estimator's
              requires_s tag is True, then an error will be raised.
            - If `'no_validation'`, `check_array` is called on `X` or `y` and the
              estimator's requires_s tag is ignored. This is a default
              placeholder and is never meant to be explicitly set. In that case
              `X` must be passed.
            - Otherwise, only `s` with `_check_s` or both `X` and `s` are
              checked with either `check_array` or `check_X_y_s` depending on
              `validate_separately`.

        reset : bool, default=True
            Whether to reset the `n_features_in_` attribute.
            If False, the input will be checked for consistency with data
            provided when reset was last True.
            .. note::
               It is recommended to call reset=True in `fit` and in the first
               call to `partial_fit`. All other methods that validate `X`
               should set `reset=False`.

        **check_params : kwargs
            Parameters passed to :func:`sklearn.utils.check_array` or
            :func:`sklearn.utils.check_X_y`. Ignored if validate_separately
            is not False.

            `estimator=self` is automatically added to these params to generate
            more informative error message in case of invalid input data.

        Returns
        -------
        out : {ndarray, sparse matrix} or tuple of these
            The validated input. A tuple is returned if `X`, `y`, and `s` are validated.
        """

        self._check_feature_names(X, reset=reset)

        if y is None and self._get_tags()["requires_y"]:
            raise ValueError(
                f"This {self.__class__.__name__} estimator requires y to be passed, but the target y is None."
            )

        if s is None and self._get_tags()["requires_s"]:
            raise ValueError(
                f"This {self.__class__.__name__} estimator requires s to be passed, but the sensitive feature s is None."
            )

        no_val_X = isinstance(X, str) and X == "no_validation"
        no_val_y = y is None or isinstance(y, str) and y == "no_validation"
        no_val_s = s is None or isinstance(s, str) and s == "no_validation"

        default_check_params = {"estimator": self}
        check_params = {**default_check_params, **check_params}

        if no_val_X and no_val_y and no_val_s:
            raise ValueError("Validation should be done on X, y, or s, or any combination among them.")
        elif not no_val_X and no_val_y and no_val_s:
            # validate only X
            X = check_array(X, input_name="X", **check_params)
            out = X
        elif no_val_X and not no_val_y and no_val_s:
            # validate only y
            y = _check_y(y, **check_params)
            out = y
        elif no_val_X and no_val_y and not no_val_s:
            # validate only s
            s = _check_y(s, **check_params)  # TODO: correct the input check_params for s
            out = s
        elif not no_val_X and not no_val_y and no_val_s:
            # validate X and y
            X, y = check_X_y(X, y, **check_params)
            out = X, y
        elif not no_val_X and no_val_y and not no_val_s:
            # validate X and s
            X, s = check_X_y(X, s, **check_params)  # TODO: correct the input check_params for s
            out = X, s
        elif no_val_X and not no_val_y and not no_val_s:
            # validate y and s
            y = _check_y(y, **check_params)
            s = _check_y(s, **check_params)  # TODO: correct the input check_params for s
            check_consistent_length(y, s)
            out = y, s
        else:
            # validate X, y, and s
            X = check_array(X, input_name="X", **check_params)
            y = _check_y(y, **check_params)
            s = _check_y(s, **check_params)  # TODO: correct the input check_params for s
            check_consistent_length(X, y, s)
            out = X, y, s

        if not no_val_X and check_params.get("ensure_2d", True):
            self._check_n_features(X, reset=reset)

        return out

    def _validate_idx(self, idx: Union[int, npt.ArrayLike], max_idx: int) -> Tuple[int, ...]:
        """ Validate feature indexes, primarily for sensitive attributes in input X """

        if isinstance(idx, int):
            idx = tuple([idx])

        if not all(isinstance(i, int) for i in idx):
            raise ValueError("Expects all indexes an integer")
        elif min(idx) < 0 or max(idx) >= max_idx:
            raise ValueError("Indexes out of range [0, feature dimensions - 1]")
        else:
            idx = tuple(i for i in sorted(set(idx)))

        return idx

    def _validate_cls_y(self, y: npt.NDArray) -> npt.NDArray[np.int16]:
        """ Validate and encode target variables for classification """

        y_type = type_of_target(y)
        if y_type not in self._get_tags()["y_types"]:
            raise ValueError(
                f"This {self.__class__.__name__} estimator does not support {y_type} targets."
            )

        enc = LabelEncoder()
        y = enc.fit_transform(y)

        self.classes_ = enc.classes_
        if len(self.classes_) < 2:
            raise ValueError(
                f"This {self.__class__.__name__} estimator need at least two type of targets, but the data only contains one: {self.classes_[0]}."
            )

        return y

    def _validate_grp_s(self, s: npt.ArrayLike) -> npt.NDArray[np.int16]:
        """ Validate and encode input sensitive features for group fairness """

        s_type = type_of_target(s)
        if s_type not in self._get_tags()["s_types"]:
            raise ValueError(
                f"This {self.__class__.__name__} estimator does not support {s_type} sensitive attributes."
            )

        enc = LabelEncoder()
        s = enc.fit_transform(s)

        self.s_classes_ = enc.classes_
        if len(self.s_classes_) < 2:
            raise ValueError(
                f"This {self.__class__.__name__} estimator need at least two sensitive groups, but the data only contains only one group: {self.s_classes_[0]}."
            )

        return s



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import numpy as np
from typing import Sequence, Union


def onehottify(x: Union[np.ndarray, Sequence], n=None, dtype=np.float32):
    """ 1-hot encode x with the max value n (computed from data if n is None) """
    if not isinstance(x, np.ndarray):
        x = np.asarray(x)
    n = np.max(x) + 1 if n is None else n
    return np.eye(n, dtype=dtype)[x]



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

from __future__ import annotations

import numpy as np
import numpy.typing as npt
from itertools import combinations, product
from numbers import Real, Integral
from scipy.special import expit as sigmoid
from typing import Any

from sklearn.utils.validation import check_is_fitted
from sklearn.utils._array_api import get_namespace
from sklearn.utils._param_validation import StrOptions, Interval

from ._base import FairEstimator


def clipped_sigmoid(z, eps=1e-6):
    return np.clip(sigmoid(z), a_min=eps, a_max=1. - eps)


class FairGLM(FairEstimator):
    """
    "Fair Generalized Linear Models with a Convex Penalty"

    Reference:
        https://proceedings.mlr.press/v162/do22a/do22a.pdf
    Code adopted from:
        https://github.com/hyungrok-do/fair-glm-cvx

    # TODO; add support to multi-class and regression

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        A list of class labels known to the classifier.

    s_classes_ : ndarray of shape (n_sensitive_group,)
        A list of sensitive classes known to the classifier.

    n_features_in_ : int
        Number of features seen during `fit`.

    feature_names_in_ : ndarray of shape (n_features_in_,)
        Names of features seen during `fit`. Defined only when X has feature names that are all strings.

    coef_ : ndarray of shape (1, n_features)
        Coefficient of the features in the decision function.

    intercept_ : ndarray of shape (1,)
        Intercept (a.k.a. bias) added to the decision function.

    Examples
    --------
    >>> from sklearn.preprocessing import StandardScaler
    >>> from fairpy.dataset import Adult
    >>> from fairpy.model import FairGLM
    >>> dataset = Adult()
    >>> split_data = dataset.split()
    >>> model = FairGLM()
    >>> model.fit(split_data.X_train, split_data.y_train, split_data.s_train)
    >>> model.predict(split_data.X_test)
    """

    _parameter_constraints = {
        "solver": [StrOptions({"CG", "newton"})],
        "fit_intercept": ["boolean"],
        "max_iter": [Interval(Integral, 0, None, closed="left")],
        "lam": [Interval(Real, 0, None, closed="left")],
        "tol": [Interval(Real, 0, None, closed="left")],
    }

    def __init__(
            self,
            solver: str = "CG",
            fit_intercept: bool = True,
            max_iter: int = 100,
            lam: float = 1e-3,
            tol: float = 1e-4,
    ):
        """
        Parameters
        ----------
        solver : {"CG", "newton"}, default="CG"
            Algorithm to use in the optimization problem.
            CG: conjugate gradient method, first-order optimization
            newton: newton-raphson method, second-order optimization

        fit_intercept : bool, default=True
            Specifies if a constant (a.k.a. bias or intercept) should be added to the decision function.

        max_iter : int, default=100
            Maximum number of iterations taken for the solvers to converge.

        lam : float, default=1e-3
            Fairness regularization strength.

        tol : float, default=1e-4
            Tolerance for stopping criteria.
        """

        self.solver = solver
        self.fit_intercept = fit_intercept
        self.max_iter = max_iter
        self.lam = lam
        self.tol = tol

    def _compute_D(self, X: npt.NDArray, y: npt.NDArray, s: npt.NDArray) -> npt.NDArray[np.float64]:
        """ Pre-compute the matrix D in equation 8 """

        n_samples, n_features = X.shape
        D = np.zeros((n_features, n_features))
        if self.lam > 0:
            for (a, b), yd in product(combinations(self.s_classes_, 2), self.classes_):  # TODO: fix s_classes
                Xay = X[np.logical_and(s == a, y == yd)]
                Xby = X[np.logical_and(s == b, y == yd)]
                diff = (Xay[None, :, :] - Xby[:, None, :]).reshape(-1, n_features)
                D += diff.T @ diff / (np.float64(len(Xay)) * np.float64(len(Xby)))

            # normalization in equation 9
            D = (self.lam * 2) / (len(self.classes_) * len(self.s_classes_) * (len(self.classes_) - 1)) * D

        return D

    def _cg(self, X: npt.NDArray, y: npt.NDArray, s: npt.NDArray) -> FairGLM:
        """ Conjugate gradient method """

        n_samples, n_features = X.shape
        beta = np.zeros(n_features)
        ls_grid = np.exp(np.linspace(np.log(1e-3), np.log(1), 20))

        D = self._compute_D(X, y, s)

        def loss_fn(b):
            xb = X @ b
            return -np.sum(y * xb - np.log(1 + np.exp(xb))) / n_samples + .5 * b @ D @ b

        grad_fn = lambda b: -X.T @ (y - clipped_sigmoid(X @ b)) / N + D @ b
        conj_fn = lambda b_n, b_o: np.max([0, b_n @ (b_n - b_o) / np.clip(b_o @ b_o, a_min=1e-4, a_max=None)])

        grad_old = grad_fn(beta)
        conj_old = np.copy(grad_old)
        vio = np.nan
        for _ in range(self.max_iter):
            grad = grad_fn(beta)

            vio = np.linalg.norm(grad)
            if vio <= self.tol:
                break

            conj = grad + conj_fn(grad, grad_old) * conj_old
            cand = [beta - conj * v for v in ls_grid]
            ls_min = np.argmin([loss_fn(c) for c in cand])
            beta = cand[ls_min]

            grad_old = np.copy(grad)
            conj_old = np.copy(conj)

        if vio > self.tol:
            import warnings
            from sklearn.exceptions import ConvergenceWarning
            warnings.warn("FairGLM does not converged, please consider increase the maximum iteration",
                          ConvergenceWarning)

        if self.fit_intercept:
            self.coef_ = beta[1:]
            self.intercept_ = beta[0]
        else:
            self.coef_ = beta
            self.intercept_ = None

        return self

    def _newton(self, X: npt.NDArray, y: npt.NDArray, s: npt.NDArray) -> FairGLM:
        """ Newton-Raphson method """

        N, P = X.shape
        beta = [np.zeros(P)]

        D = self._compute_D(X, y, s)

        vio = np.nan
        for i in range(self.max_iter):
            mu = 1. / (1 + np.exp(np.clip(-np.dot(X, beta[i]), -50., 50.)))
            grad = -X.T.dot(y - mu) / N + np.dot(D, beta[i])

            w = np.diag(mu * (1 - mu))
            try:
                hinv = np.linalg.inv(X.T.dot(w).dot(X) / N + D)
            except:
                raise ValueError("Matrix not invertible in newton method, consider switching solver to 'CG'.")

            beta.append(beta[i] - np.dot(hinv, grad))

            vio = np.linalg.norm(grad)
            if vio <= self.tol:
                break

        if vio > self.tol:
            import warnings
            from sklearn.exceptions import ConvergenceWarning
            warnings.warn("FairGLM does not converged, please consider increase the maximum iteration",
                          ConvergenceWarning)

        if self.fit_intercept:
            self.coef_ = beta[-1][1:]
            self.intercept_ = beta[-1][0]
        else:
            self.coef_ = beta[-1]
            self.intercept_ = None

        return self

    def fit(self, X: npt.ArrayLike, y: npt.ArrayLike, s: npt.ArrayLike):
        """
        Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training vector, where `n_samples` is the number of samples and `n_features` is the number of features.

        y : array-like of shape (n_samples,)
            Target vector relative to X.

        s : array-like of shape (n_samples,)
            Sensitive attributes relative to X.

        Returns
        -------
        self
            Fitted estimator.
        """

        self._validate_params()
        X, y, s = self._validate_data(X, y, s)
        y = self._validate_cls_y(y)
        s = self._validate_grp_s(s)

        if self.fit_intercept:
            X = np.column_stack([np.ones(len(X)), X])

        if self.solver == "CG":
            return self._cg(X, y, s)
        elif self.solver == "newton":
            return self._newton(X, y, s)
        else:
            raise NotImplementedError

    def decision_function(self, X: npt.ArrayLike) -> npt.NDArray[np.float64]:
        """
        Predict confidence scores for samples.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data matrix for which we want to get the confidence scores.

        Returns
        -------
        scores : ndarray of shape (n_samples,)
            Confidence scores per `(n_samples, n_classes)` combination. In the
            binary case, confidence score for `self.classes_[1]` where >0 means
            this class would be predicted.
        """

        check_is_fitted(self)
        X = self._validate_data(X, reset=False)

        if self.fit_intercept:
            scores = np.dot(X, self.coef_) + self.intercept_
        else:
            scores = np.dot(X, self.coef_)
        scores = 1. / (1. + np.exp(-scores))

        return scores

    def predict(self, X: npt.ArrayLike) -> npt.NDArray[Any]:
        """
        Predict class labels for samples in X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data matrix for which we want to get the predictions.

        Returns
        -------
        y_pred : ndarray of shape (n_samples,)
            Vector containing the class labels for each sample.
        """

        xp, _ = get_namespace(X)
        scores = self.decision_function(X)
        indices = xp.astype(scores > 0, int)

        return xp.take(self.classes_, indices, axis=0)

    def predict_proba(self, X: npt.ArrayLike) -> npt.NDArray[np.float64]:
        """
        Probability estimates.

        The returned estimates for all classes are ordered by the
        label of classes.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Vector to be scored, where `n_samples` is the number of samples and
            `n_features` is the number of features.

        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Returns the probability of the sample for each class in the model,
            where classes are ordered as they are in ``self.classes_``.
        """

        prob = self.decision_function(X)
        prob = sigmoid(prob)

        return np.vstack([1 - prob, prob]).T

    def _more_tags(self):
        return {
            'binary_only': True,
            'requires_y': True,
            'requires_s': True,
            "y_types": ["binary"],
            "s_types": ["binary"],
        }



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

from __future__ import annotations

from numbers import Real, Integral
from typing import Any

import numpy as np
import numpy.typing as npt

from sklearn.utils.validation import check_is_fitted
from sklearn.utils.multiclass import type_of_target
from sklearn.linear_model import LogisticRegression
from sklearn.utils._param_validation import StrOptions, HasMethods, Interval

from ._base import FairEstimator
from ..utils.encode import onehottify


class LabelBias(FairEstimator):
    """
    "Identifying and Correcting Label Bias in Machine Learning"
        Adaptively learn the weights for sensitive groups by fitting the sub-estimator multiple times

    Reference:
        http://proceedings.mlr.press/v108/jiang20a/jiang20a.pdf
    Code adopted from:
        https://github.com/google-research/google-research/tree/master/label_bias

    # TODO: add support for equal odds

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        A list of class labels known to the classifier.

    s_classes_ : ndarray of shape (n_sensitive_group,)
        A list of sensitive classes known to LabelBias during training.

    n_features_in_ : int
        Number of features seen during `fit`.

    feature_names_in_ : ndarray of shape (n_features_in_,)
        Names of features seen during `fit`. Defined only when X has feature names that are all strings.

    weights_ : ndarray of shape (n_sample,)
        Weights for training samples solved by LabelBias.

    fitted_estimator_ : an object of the classifier
        Fitted estimator.

    Examples
    --------
    >>> from sklearn.preprocessing import StandardScaler
    >>> from fairpy.dataset import Adult
    >>> from fairpy.model import LabelBias
    >>> dataset = Adult()
    >>> split_data = dataset.split()
    >>> model = LabelBias()
    >>> model.fit(split_data.X_train, split_data.y_train, split_data.s_train)
    >>> model.predict(split_data.X_test)
    """

    _parameter_constraints = {
        "metric": [StrOptions({"dp", "eop"})],
        "estimator": [HasMethods(["fit", "predict"]), None],
        "max_iter": [Interval(Integral, 0, None, closed="left")],
        "tol": [Interval(Real, 0, None, closed="left")],
        "lr": [Interval(Real, 0, None, closed="left")],
    }

    def __init__(
            self,
            metric: str = "dp",
            estimator: Any = None,
            max_iter: int = 100,
            tol: float = 1e-3,
            lr: float = 1.
    ):
        """
        Parameters
        ----------
        metric : {"dp", "eop"}, default="dp"
            The fairness notion adopted in optimization.
            dp: demographic parity
            eop: equal opportunity

        estimator : a callable object with some requirements, default=None
            Estimator is the base classifier, and LabelBias is a wrapper function over this classifier.
            The estimator should implement 'fit' and 'predict' methods (and 'predict_proba' if needed) for
            model's training and prediction. 'fit' should accept arguments: X, y, sample_weight. 'predict'
            should accept argument X. If not specific, use Logistic Regression as the estimator.

        max_iter : int, default=100
            Maximum number of iterations taken for the solvers to converge.

        tol : float, default=1e-3
            Tolerance for stopping criteria. It operates on the maximum value of violation.

        lr : float, default=1
            The learning rate for multipliers' updates.
        """

        self.metric = metric
        self.estimator = estimator
        self.max_iter = max_iter
        self.tol = tol
        self.lr = lr

    def _debias_weights(self, y: npt.NDArray, s_one_hot: npt.NDArray, multipliers: npt.NDArray) -> npt.NDArray:
        """ Update instance's weight based on multipliers """

        exponents = np.zeros(y.shape[0], dtype=np.float32)
        exponents = exponents - np.sum(np.multiply(s_one_hot, multipliers[np.newaxis, :]), axis=1)
        weights = np.exp(exponents) / (np.exp(exponents) + np.exp(-exponents))
        weights = np.where(y > 0, 1 - weights, weights)

        return weights

    def _dp_vio(self, y_pred: npt.NDArray, y: npt.NDArray, s_one_hot: npt.NDArray) -> npt.NDArray:
        """ Violation in demographic parity """

        vio = []
        base = np.mean(y_pred)
        for i in range(s_one_hot.shape[1]):
            idx = np.where(s_one_hot[:, i] == 1)
            vio.append(base - np.mean(np.take(y_pred, idx)))
        vio = np.asarray(vio)

        return vio

    def _eop_vio(self, y_pred: npt.NDArray, y: npt.NDArray, s_one_hot: npt.NDArray) -> npt.NDArray:
        """ Violation in equal opportunity """

        vio = []
        pos_idx = np.where(y == 1)
        base = np.mean(np.take(y_pred, pos_idx))
        for i in range(s_one_hot.shape[1]):
            idx = np.intersect1d(np.where(s_one_hot[:, i] == 1), pos_idx)
            vio.append(base - np.mean(np.take(y_pred, idx)))
        vio = np.asarray(vio)

        return vio

    def fit(self, X: npt.ArrayLike, y: npt.ArrayLike, s: npt.ArrayLike) -> LabelBias:
        """
        Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training vector, where `n_samples` is the number of samples and `n_features` is the number of features.

        y : array-like of shape (n_samples,)
            Target vector relative to X.

        s : array-like of shape (n_samples,)
            Sensitive attributes relative to X.

        Returns
        -------
        self
            Fitted estimator.
        """

        self._validate_params()
        X, y, s = self._validate_data(X, y, s)
        y = self._validate_cls_y(y)
        s = self._validate_grp_s(s)

        if self.estimator is None:
            model = LogisticRegression()
        else:
            model = self.estimator

        s_one_hot = onehottify(s)

        if self.metric == "dp":
            _vio_func = self._dp_vio
        elif self.metric == "eop":
            _vio_func = self._eop_vio
        else:
            raise NotImplementedError

        model.fit(X, y, sample_weight=self.weights_)
        y_pred = model.predict(X)
        y_pred_type = type_of_target(y_pred)
        if y_pred_type not in self._get_tags()["y_types"]:
            raise ValueError(
                f"This {self.__class__.__name__} estimator does not support {y_pred_type} predictions from estimator."
            )

        multipliers = np.zeros(len(self.s_classes_))
        self.weights_ = np.ones(X.shape[0])

        vio = np.nan
        index_func = lambda x: self.classes_.index(x)
        for _ in range(self.max_iter):
            self.weights_ = self._debias_weights(y, s_one_hot, multipliers)
            model.fit(X, y, self.weights_)
            y_pred = model.predict(X)
            y_pred = np.fromiter(map(index_func, y_pred), dtype=np.int16)
            vio = _vio_func(y_pred, y, s_one_hot)
            if np.max(vio) <= self.tol:
                break
            else:
                multipliers += vio * self.lr

        self.fitted_estimator_ = model

        if np.max(vio) > self.tol:
            import warnings
            from sklearn.exceptions import ConvergenceWarning
            warnings.warn("LabelBias does not converged, please consider increase the maximum iteration",
                          ConvergenceWarning)

        return self

    def predict(self, X: npt.ArrayLike) -> npt.NDArray[Any]:
        """
        Predict class labels for samples in X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data matrix for which we want to get the predictions.

        Returns
        -------
        y_pred : ndarray of shape (n_samples,)
            Vector containing the class labels for each sample.
        """

        check_is_fitted(self)
        X = self._validate_data(X, reset=False)

        return self.fitted_estimator_.predict(X)

    def predict_proba(self, X: np.ndarray) -> Any:
        """
        Probability estimates. Only available if 'predict_proba' is implemented in estimator.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Vector to be scored, where `n_samples` is the number of samples and
            `n_features` is the number of features.

        Returns
        -------
        T : array-like of shape (n_samples, n_classes)
            Returns the probability of the sample for each class in the model,
            where classes are ordered as they are in ``self.classes_``.
        """

        if not hasattr(self.estimator, "predict_proba"):
            raise ValueError("Estimator does not have method 'predict_proba'")

        check_is_fitted(self)
        X = self._validate_data(X, reset=False)

        return self.fitted_estimator_.predict_proba(X)

    def _more_tags(self):
        return {
            'binary_only': True,
            'requires_y': True,
            'requires_s': True,
            "y_types": ["binary"],
            "s_types": ["binary", "multiclass"],
        }



from setuptools import setup, find_packages

setup(
    name='fairpy',
    packages=find_packages(),
    version='0.1.0',
    description='FairPy is a comprehensive Python Library for Machine Learning Fairness',
    url='https://github.com/brandeis-machine-learning/FairPy',
    license='MIT',
    classifiers=[
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    ],
)


# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import os
import sys
import pandas as pd
from typing import Sequence, Union

from ._base import TabDataset


class Compas(TabDataset):
    """ https://github.com/propublica/compas-analysis """

    file2url = {
        "compas-scores-two-years.csv": "https://raw.githubusercontent.com/propublica/compas-analysis/master/compas-scores-two-years.csv",
    }

    features_to_keep = (
        "sex",
        "age",
        "age_cat",
        "race",
        "juv_fel_count",
        "juv_misd_count",
        "juv_other_count",
        "priors_count",
        "c_charge_degree",
        "c_charge_desc",
        "two_year_recid",
    )

    categorical_feat = [
        "sex",
        "age_cat",
        "race",
        "c_charge_degree",
        "c_charge_desc",
    ]

    numerical_feat = [
        "age",
        "juv_fel_count",
        "juv_misd_count",
        "juv_other_count",
        "priors_count",
    ]

    drop_feat = []
    na_values = ()

    label_name = "two_year_recid"
    label_mapping = ("Did recid.", "No recid.")

    dataset_name = "compas"

    def __init__(
            self,
            sen_feat: Union[str, Sequence[str]] = "race",
            dir_path: str = os.path.join(sys.path[1], "fairpy/data/compas"),
            download: bool = False,
    ):
        self._check_and_download_data(dir_path, self.file2url, download)
        df = pd.read_csv(os.path.join(dir_path, "compas-scores-two-years.csv"), index_col='id')

        super(Compas, self).__init__(
            df=df,
            label_name=self.label_name,
            sen_feat=sen_feat,
            cat_feat=self.categorical_feat,
            num_feat=self.numerical_feat,
            dataset_name=self.dataset_name,
            custom_preprocessing=self.custom_preprocessing,
            na_values=self.na_values,
            drop_feat=self.drop_feat,
            label_mapping=self.label_mapping,
        )

    def custom_preprocessing(self, df):
        """
        Perform the same preprocessing as the original analysis:
        https://github.com/propublica/compas-analysis/blob/master/Compas%20Analysis.ipynb
        """

        def two_year_recid(row):
            return 'Did recid.' if row['two_year_recid'] == 1 else 'No recid.'

        df['two_year_recid'] = df.apply(lambda row: two_year_recid(row), axis=1)
        df = df[
            (df.days_b_screening_arrest <= 30)
            & (df.days_b_screening_arrest >= -30)
            & (df.is_recid != -1)
            & (df.c_charge_degree != 'O')
            & (df.score_text != 'N/A')
            ]
        df = df[list(self.features_to_keep)]

        return df



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import os
import sys
import pandas as pd
from typing import Sequence, Union

from ._base import TabDataset


class Dutch(TabDataset):
    """ https://github.com/tailequy/fairness_dataset/blob/main/experiments/data/dutch.csv """

    file2url = {
        "dutch.csv": "https://raw.githubusercontent.com/tailequy/fairness_dataset/main/experiments/data/dutch.csv",
    }

    categorical_feat = [
        "sex",
        "age",
        "household_position",
        "household_size",
        "prev_residence_place",
        "citizenship",
        "country_birth",
        "edu_level",
        "economic_status",
        "cur_eco_activity",
        "marital_status",
    ]

    numerical_feat = [
    ]

    drop_feat = []
    na_values = ()

    label_name = "occupation"
    label_mapping = (0, 1)

    dataset_name = "dutch"

    def __init__(
            self,
            sen_feat: Union[str, Sequence[str]] = "sex",
            dir_path: str = os.path.join(sys.path[1], "fairpy/data/dutch"),
            download: bool = False,
    ):
        self._check_and_download_data(dir_path, self.file2url, download)
        df = pd.read_csv(os.path.join(dir_path, "dutch.csv"))

        super(Dutch, self).__init__(
            df=df,
            label_name=self.label_name,
            sen_feat=sen_feat,
            cat_feat=self.categorical_feat,
            num_feat=self.numerical_feat,
            dataset_name=self.dataset_name,
            custom_preprocessing=None,
            na_values=self.na_values,
            drop_feat=self.drop_feat,
            label_mapping=self.label_mapping,
        )



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

import os
import sys
import pandas as pd
from typing import Sequence, Union

from ._base import TabDataset


class Bank(TabDataset):
    """ https://archive.ics.uci.edu/ml/datasets/bank+marketing """

    file2url = {
        "bank-additional.zip": "https://archive.ics.uci.edu/ml/machine-learning-databases/00222/bank-additional.zip",
    }

    categorical_feat = [
        "job",
        "marital",
        "education",
        "default",
        "housing",
        "loan",
        "contact",
        "month",
        "day_of_week",
        "poutcome",
    ]

    numerical_feat = [
        "age",
        "duration",
        "campaign",
        "pdays",
        "previous",
        "emp.var.rate",
        "cons.price.idx",
        "cons.conf.idx",
        "euribor3m",
        "nr.employed"
    ]

    drop_feat = []
    na_values = ("unknown")

    label_name = "y"
    label_mapping = ("no", "yes")

    dataset_name = "bank"

    def __init__(
            self,
            sen_feat: Union[str, Sequence[str]] = "marital",
            dir_path: str = os.path.join(sys.path[1], "fairpy/data/bank"),
            download: bool = False,
    ):
        self._check_and_download_data(dir_path, self.file2url, download)
        df = pd.read_csv(os.path.join(dir_path, "bank-additional/bank-additional-full.csv"), sep=";")

        super(Bank, self).__init__(
            df=df,
            label_name=self.label_name,
            sen_feat=sen_feat,
            cat_feat=self.categorical_feat,
            num_feat=self.numerical_feat,
            dataset_name=self.dataset_name,
            na_values=self.na_values,
            drop_feat=self.drop_feat,
            label_mapping=self.label_mapping,
        )



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

from ._label_bias import LabelBias
from ._fair_erm import LinearFairERM
from ._fair_cstr import FairCstr
from ._di_remover import DIRemover
from ._eq_odds_calib import EqOddsCalib
from ._fair_glm import FairGLM
from ._ifair import IFair
from ._fair_rank import FairRank

__all__ = [
    "LabelBias",
    "LinearFairERM",
    "FairCstr",
    "DIRemover",
    "EqOddsCalib",
    "FairGLM",
    "IFair",
    "FairRank",
]



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

from ._adult import Adult
from ._compas import Compas
from ._bank import Bank
from ._credit import Credit
from ._german import German
from ._oulad import Oulad
from ._dutch import Dutch
from ._xing import Xing

__all__ = [
    "Adult",
    "Compas",
    "Bank",
    "Credit",
    "German",
    "Oulad",
    "Dutch",
    "Xing",
]



# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

from __future__ import annotations

from copy import deepcopy
import numpy as np
import numpy.typing as npt
from numbers import Integral, Real
from typing import Any, Dict

import scipy
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.extmath import log_logistic
from sklearn.utils._param_validation import StrOptions, Interval
from sklearn.utils._array_api import get_namespace

from ._base import FairEstimator
from ..utils.encode import onehottify


class FairCstr(FairEstimator):
    """
    Fairness Constraints: Mechanisms for Fair Classification

    Currently do not have the functionality 'sep_constraint' in the original implementation.

    Reference:
        http://proceedings.mlr.press/v54/zafar17a/zafar17a.pdf
    Code adopted from:
        https://github.com/mbilalzafar/fair-classification

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,)
        A list of class labels known to the classifier.

    s_classes_ : ndarray of shape (n_sensitive_group,)
        A list of sensitive classes known to LabelBias during training.

    n_features_in_ : int
        Number of features seen during `fit`.

    feature_names_in_ : ndarray of shape (n_features_in_,)
        Names of features seen during `fit`. Defined only when X has feature names that are all strings.

    coef_ : ndarray of shape (1, n_features)
        Coefficient of the features in the decision function.

    Examples
    --------
    >>> from sklearn.preprocessing import StandardScaler
    >>> from fairpy.dataset import Adult
    >>> from fairpy.model import FairCstr
    >>> dataset = Adult()
    >>> split_data = dataset.split()
    >>> model = FairCstr()
    >>> model.fit(split_data.X_train, split_data.y_train, split_data.s_train)
    >>> model.predict(split_data.X_test)
    """

    _parameter_constraints = {
        "cstr": [StrOptions({"fair", "acc"})],
        "sep_cstr": ["boolean"],
        "gamma": [Interval(Real, 0, None, closed="left")],
        "max_iter": [Interval(Integral, 0, None, closed="left")],
    }

    def __init__(self, cstr: str = "fair", sep_cstr: bool = False, gamma: float = 0.5, max_iter: int = 100000):
        """
        Parameters
        ----------
        cstr : {"fair", "acc"}, default="fair"
            Constraints added to the optimization.
            fair: optimize accuracy subject to fairness constraint
            acc: optimize fairness subject to accuracy constraint

        sep_cstr : bool, default=False
            If set to True, separate the constraints for different instances when cstr="acc".

        gamma : float, default=0.5
            Strength in fairness when cstr="acc". A larger gamma means more loss in accuracy in achieving fairness.

        max_iter : int, default=100
            Maximum number of iterations taken for the solvers to converge.

        """

        self.cstr = cstr
        self.sep_cstr = sep_cstr
        self.gamma = gamma
        self.max_iter = max_iter

    def _logistic_loss(self, w: npt.NDArray, X: npt.NDArray, y: npt.NDArray, return_arr: bool = False) -> npt.NDArray:
        """ Compute the logistic loss """

        yz = y * np.dot(X, w)
        if return_arr:
            out = -(log_logistic(yz))
        else:
            out = -np.sum(log_logistic(yz))
        return out

    def _s_cov(self, w: npt.NDArray, X: npt.NDArray, y: npt.NDArray, s_one_col: npt.NDArray, thresh: float = 0.):
        """ Constraint on fairness relative to sensitive feature when cstr="fair" """

        if w is None:
            arr = y
        else:
            arr = np.dot(w, X.T)

        cov = np.dot(s_one_col - np.mean(s_one_col), arr) / len(s_one_col)
        return thresh - abs(cov)

    def _get_cstr_cov(self, X: npt.NDArray, y: npt.NDArray, s_one_col: npt.NDArray) -> Dict:
        constraints = {"type": "ineq", "fun": self._s_cov, "args": (X, y, s_one_col)}
        return constraints

    def _cstr_protected(self, w: npt.NDArray, x: npt.NDArray) -> float:
        """ Constraint on instance that should not be misclassified to negative class when cstr="acc" """
        return np.dot(w, x)

    def _cstr_unprotected(self, w: npt.NDArray, old_loss: npt.NDArray, x: npt.NDArray, y: npt.NDArray) -> float:
        new_loss = self._logistic_loss(w, x, y)
        return ((1.0 + self.gamma) * old_loss) - new_loss

    def _cstr_gamma_all(self, w: npt.NDArray, X: npt.NDArray, y: npt.NDArray, initial_loss_arr: npt.NDArray) -> float:
        new_loss = self._logistic_loss(w, X, y)
        old_loss = sum(initial_loss_arr)
        return ((1.0 + self.gamma) * old_loss) - new_loss

    def _cross_cov_abs_func(self, w: npt.NDArray, X: npt.NDArray, s: npt.NDArray) -> float:
        cross_cov = (s - np.mean(s)) * np.dot(w, X.T)
        return float(abs(sum(cross_cov))) / float(X.shape[0])

    def fit(self, X: npt.ArrayLike, y: npt.ArrayLike, s: npt.ArrayLike) -> FairCstr:
        """
        Fit the model according to the given training data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training vector, where `n_samples` is the number of samples and `n_features` is the number of features.

        y : array-like of shape (n_samples,)
            Target vector relative to X.

        s : array-like of shape (n_samples,)
            Sensitive attributes relative to X.

        Returns
        -------
        self
            Fitted estimator.
        """

        self._validate_params()
        X, y, s = self._validate_data(X, y, s)
        y = self._validate_cls_y(y)
        s = self._validate_grp_s(s)
        s_one_hot = onehottify(s)

        if self.cstr == "fair":
            # constraint on fairness, support multi-value sensitive attributes

            constraints = []
            for i in range(s_one_hot.shape[1]):
                constraints.append(self._get_cstr_cov(X, y, s_one_hot[:, i]))

            w = scipy.optimize.minimize(
                fun=self._logistic_loss,
                x0=np.random.rand(X.shape[1], ),
                args=(X, y),
                method='SLSQP',
                options={"maxiter": self.max_iter},
                constraints=constraints,
            )

        elif self.cstr == "acc":
            # constraint on accuracy, only support binary sensitive attribute

            w = scipy.optimize.minimize(
                fun=self._logistic_loss,
                x0=np.random.rand(X.shape[1], ),
                args=(X, y),
                method='SLSQP',
                options={"maxiter": self.max_iter},
            )
            old_w = deepcopy(w.x)

            constraints = []
            predicted_labels = np.sign(np.dot(w.x, X.T))
            unconstrained_loss_arr = self._logistic_loss(w.x, X, y, return_arr=True)

            if self.sep_cstr:
                for i in range(len(predicted_labels)):
                    if predicted_labels[i] == 1. and s[i] == 1.:
                        constraints.append({'type': 'ineq', 'fun': self._cstr_protected, 'args': (X[i])})
                    else:
                        constraints.append({
                            'type': 'ineq',
                            'fun': self._cstr_unprotected,
                            'args': (i, unconstrained_loss_arr[i], X[i], y[i])
                        })
            else:
                constraints.append(
                    {'type': 'ineq', 'fun': self._cstr_gamma_all, 'args': (X, y, unconstrained_loss_arr)})

            w = scipy.optimize.minimize(
                fun=self._cross_cov_abs_func,
                x0=old_w,
                args=(X, s),
                method='SLSQP',
                options={"maxiter": self.max_iter},
                constraints=constraints,
            )

        else:
            raise NotImplementedError

        self.coef_ = w.x

        return self

    def decision_function(self, X: npt.ArrayLike) -> npt.NDArray[np.float64]:
        """
        Predict confidence scores for samples.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data matrix for which we want to get the confidence scores.

        Returns
        -------
        scores : ndarray of shape (n_samples,)
            Confidence scores per `(n_samples, n_classes)` combination. In the
            binary case, confidence score for `self.classes_[1]` where >0 means
            this class would be predicted.
        """

        check_is_fitted(self)
        X = self._validate_data(X, reset=False)

        scores = np.dot(X, self.coef_)
        # TODO: probably not correct, verify through sklearn?
        scores = 1. / (1. + np.exp(-scores))

        return scores

    def predict(self, X: npt.ArrayLike) -> npt.NDArray[Any]:
        """
        Predict class labels for samples in X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data matrix for which we want to get the predictions.

        Returns
        -------
        y_pred : ndarray of shape (n_samples,)
            Vector containing the class labels for each sample.
        """

        xp, _ = get_namespace(X)
        scores = self.decision_function(X)
        indices = xp.astype(scores > 0, int)

        return xp.take(self.classes_, indices, axis=0)

    def _more_tags(self):
        return {
            'binary_only': True,
            'requires_y': True,
            'requires_s': True,
            "y_types": ["binary"],
            "s_types": ["binary"],
        }
